#include "cfg.h"

#ifdef GRDRV_GEM

#include "links.h"

#include <mintbind.h>
#include <sysvars.h>
#include <gem.h>
#include <string.h>

#undef DEBUG_GEM

#define HACK_TEXT    /* GEM fonts if dither_letters at 0 */
#define HACK_BUTTONS /* Lynx buttons => GEM 3D buttons */

#define MAX_DEV   8
#define MAX_PATH  1024

#define FORM1     0  /* buttons bar, ilinks.rsc */
#define BBOX      0
#define BBACK     1
#define BNEXT     2
#define BSTOP     3
#define BRELOAD   4
#define BURL      5
#define BLIST     6

#define DD_OK         0 /* Drag & Drop */
#define DD_NAK        1
#define DD_EXT        2
#define DD_LEN        3
#define DD_TRASH      4
#define DD_PRINTER    5
#define DD_CLIPBOARD  6

#define DD_TIMEOUT    1000
#define DD_NUMEXTS    8
#define DD_EXTSIZE    32L
#define DD_NAMEMAX    128
#define DD_HDRMAX     (8+DD_NAMEMAX+DD_NAMEMAX)

#define VA_START      0x4711
#define AV_STARTED    0x4738

extern struct graphics_driver gem_driver;

static short ap_id, gr_hwchar, gr_hhchar, desk_x, desk_y, desk_w, desk_h, w_border_3d, h_border_3d;
static short vdi_handle, width_screen, height_screen;
#ifdef HACK_TEXT
static short gem_fonts;
#endif
static short bpp, pixelsize, packed_format, little_endian;
static struct graphics_device *gem_dev[MAX_DEV];
static char path_scrp[MAX_PATH];
static unsigned char tab_index_col_png[256];
static long tab_rgb[256];
static short save_palette[256 * 3];
OBJECT *buttons = NULL;
static unsigned char *icon;

struct gem_window
{
	short handle;
	short type;
	short x, y, w, h;
	struct { short x, y, w, h; } redraw;
	struct { short x, y, w, h; } clip;
	int buttons;
	int fulled;
	int flag_redraw;
	int iconified;
};

#ifdef HACK_TEXT
struct gem_fontinfo
{
	struct gem_fontinfo *next;
	short id;
	short vector;
	short monospace;
	long h_factor;
	long w_factor;
	int attr;
	int default_font;
	char *name;
};

static struct gem_fontinfo *gem_pfont = NULL;
#endif	

typedef struct
{
	long ident;
	union
	{
		long l;
		short i[2];
		char c[4];
	} v;
} COOKIE;

static long gem_get_color(int search_rgb);

/* Convert a chunky screen to bitplane screen */
extern void Atari_C2pConvert4(
	unsigned char *src,     /* Source screen (one byte=one pixel) */
	unsigned char *dest,    /* Destination (4 bits planes) */
	unsigned long width,    /* Dimensions of screen to convert */
	unsigned long height,
	unsigned long dblligne, /* Double the lines when converting ? */
	unsigned long srcpitch, /* Length of one source line in bytes */
	unsigned long dstpitch  /* Length of one destination line in bytes */
);
extern void Atari_C2pConvert8(
	unsigned char *src,     /* Source screen (one byte=one pixel) */
	unsigned char *dest,    /* Destination (8 bits planes) */
	unsigned long width,    /* Dimensions of screen to convert */
	unsigned long height,
	unsigned long dblligne, /* Double the lines when converting ? */
	unsigned long srcpitch, /* Length of one source line in bytes */
	unsigned long dstpitch  /* Length of one destination line in bytes */
);

extern unsigned short Atari_table_c2p[256*4];

static TEDINFO rs_tedinfo[] = {
	"", "", "", SMALL, 0, 2, 0x1180, 0, 1, 1, 1 };
static OBJECT rs_object[] = {
	-1, -1, -1, G_BOXTEXT, OF_SELECTABLE | OF_FL3DIND, OS_NORMAL, 0, 0, 0, 0, 0 };

static const unsigned char tab_index_png_vdi[16] = { // index bitmap png color => index VDI color
	1, 4, 9, 10, 11, 8, 3, 5, 2, 7, 12, 13, 14, 15, 6, 0 };
static const unsigned char tab_index_col_pal[16] = { // index VDI color => index palette
	0, 15, 1, 2, 4, 6, 3, 5, 7, 8, 9, 10, 12, 14, 11, 13 };
static const int reduce_palette[8] = { 0, 0, 2, 2, 5, 5, 7, 7 };

static const struct {
  unsigned int 	 width;
  unsigned int 	 height;
  unsigned int 	 bytes_per_pixel; /* 3:RGB, 4:RGBA */
  unsigned char	 pixel_data[32 * 32 * 3];
} links_image = {
  32, 32, 3, 
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,000,000,255,000,000,255,000,000,
  255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,
  255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,255,255,
  255,255,255,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,255,255,255,255,255,255,255,000,000,
  255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,
  255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,
  255,000,000,255,255,255,255,255,255,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,255,255,255,
  255,255,255,255,000,000,255,000,000,255,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,255,255,255,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,255,255,255,255,255,255,255,000,000,255,000,000,255,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  255,255,255,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,255,255,255,255,255,255,255,000,000,
  255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,255,000,000,255,000,000,
  255,000,000,255,255,255,255,255,255,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,255,255,255,
  255,255,255,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,
  255,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  255,000,000,255,000,000,255,000,000,255,255,255,255,255,255,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,255,255,255,255,255,255,255,000,000,255,000,000,255,000,000,
  255,000,000,255,000,000,255,000,000,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,255,000,000,255,000,000,255,000,000,255,255,255,
  255,255,255,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,255,255,255,255,255,255,255,000,000,
  255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,255,000,000,255,000,000,
  255,000,000,255,255,255,255,255,255,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,255,255,255,
  255,255,255,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,
  255,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  255,000,000,255,000,000,255,000,000,255,255,255,255,255,255,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,255,255,255,255,255,255,255,000,000,255,000,000,255,000,000,
  255,000,000,255,000,000,255,000,000,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,255,000,000,255,000,000,255,000,000,255,255,255,
  255,255,255,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,255,255,255,255,255,255,255,000,000,
  255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,255,000,000,255,000,000,
  255,000,000,255,255,255,255,255,255,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,255,255,255,
  255,255,255,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,
  255,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  255,000,000,255,000,000,255,000,000,255,255,255,255,255,255,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,255,255,255,255,255,255,255,000,000,255,000,000,255,000,000,
  255,000,000,255,000,000,255,000,000,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,255,000,000,255,000,000,255,000,000,255,255,255,
  255,255,255,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,255,255,255,255,255,255,255,000,000,
  255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,255,000,000,255,000,000,
  255,000,000,255,255,255,255,255,255,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,
  000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,000,000,255,
  000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,255,000,000,255,
  000,000,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,
  000,000,255,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  000,000,255,000,000,255,000,000,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,000,000,255,000,000,255,000,000,255,
  000,000,255,000,000,255,000,000,255,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,255,000,000,255,000,000,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,000,000,000,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,000,000,255,
  000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,255,000,000,255,
  000,000,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,000,000,000,
  000,000,000,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,
  000,000,255,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  000,000,255,000,000,255,000,000,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,000,000,000,000,000,000,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,000,000,255,000,000,255,000,000,255,
  000,000,255,000,000,255,000,000,255,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,255,000,000,255,000,000,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,000,000,000,000,000,000,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,000,000,255,
  000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,255,000,000,255,
  000,000,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,000,000,000,000,000,000,
  000,000,000,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,
  000,000,255,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  000,000,255,000,000,255,000,000,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  000,000,000,000,000,000,000,000,000,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,000,000,255,000,000,255,000,000,255,
  000,000,255,000,000,255,000,000,255,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,255,000,000,255,000,000,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,000,000,000,000,000,000,000,000,000,000,000,000,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,000,000,255,
  000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,255,000,000,255,
  000,000,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  000,000,000,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,000,000,255,000,000,255,000,000,255,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,000,000,255,000,000,255,000,000,255,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,
  000,000,000,000,000,000,000,000,000,000,000,000,000,000,000,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,000,000,255,
  000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,
  000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,
  000,000,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,
  000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,000,000,255,
  000,000,255,000,000,255,000,000,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255
};

struct s_uni_trans {
	unsigned short unicode;
	unsigned char atari;
};

static const struct s_uni_trans Unicode_to_Atari[] = {
	{0x0080,''}, {0x0082,'\\'}, {0x0083,''}, {0x0084, 34}, {0x0085,'.'},
	{0x0086,''}, {0x0087,''}, {0x0088,'^'}, {0x0089,''}, {0x008A,'S'},
	{0x008B,'<'}, {0x008C,''}, {0x008E,'Z'}, {0x0091,'\\'}, {0x0092,'\\'},
	{0x0093, 34}, {0x0094, 34}, {0x0095,''}, {0x0096,'-'}, {0x0097,'-'},
	{0x0098,'~'}, {0x0099,''}, {0x009A,'s'}, {0x009B,'>'}, {0x009C,''},
	{0x009E,'z'}, {0x009F,'Y'}, {0x00A0,' '}, {0x00A1,''}, {0x00A2,''},
	{0x00A3,''}, {0x00A5,''}, {0x00A6,17 }, {0x00A7,''}, {0x00A8,''},
	{0x00A9,''}, {0x00AA,''}, {0x00AB,''}, {0x00AC,''}, {0x00AD,'-'},
	{0x00AE,''}, {0x00AF,''}, {0x00B0,''}, {0x00B1,''}, {0x00B2,''},
	{0x00B3,''}, {0x00B4,''}, {0x00B5,''}, {0x00B6,''}, {0x00B7,''},
	{0x00B9,'\\'}, {0x00BA,''}, {0x00BB,''}, {0x00BC,''}, {0x00BD,''},
	{0x00BE,' '}, {0x00BF,''}, {0x00C0,''}, {0x00C1,'A'}, {0x00C2,'A'},
	{0x00C3,''}, {0x00C4,''}, {0x00C5,''}, {0x00C6,''}, {0x00C7,''},
	{0x00C8,'E'}, {0x00C9,''}, {0x00CA,'E'}, {0x00CB,'E'}, {0x00CC,'I'},
	{0x00CD,'I'}, {0x00CE,'I'}, {0x00CF,'I'}, {0x00D0,'D'}, {0x00D1,''},
	{0x00D2,'O'}, {0x00D3,'O'}, {0x00D4,'O'}, {0x00D5,''}, {0x00D6,''},
	{0x00D7,'x'}, {0x00D8,''}, {0x00D9,'U'}, {0x00DA,'U'}, {0x00DB,'U'},
	{0x00DC,''}, {0x00DD,'Y'}, {0x00DF,''}, {0x00E0,''}, {0x00E1,''},
	{0x00E2,''}, {0x00E3,''}, {0x00E4,''}, {0x00E5,''}, {0x00E6,''},
	{0x00E7,''}, {0x00E8,''}, {0x00E9,''}, {0x00EA,''}, {0x00EB,''},
	{0x00EC,''}, {0x00ED,''}, {0x00EE,''}, {0x00EF,''}, {0x00F0,''},
	{0x00F1,''}, {0x00F2,''}, {0x00F3,''}, {0x00F4,''}, {0x00F5,''},
	{0x00F6,''}, {0x00F7,''}, {0x00F8,''}, {0x00F9,''}, {0x00FA,''},
	{0x00FB,''}, {0x00FC,''}, {0x00FD,'y'}, {0x00FF,''}, {0x0108,'C'},
	{0x0109,'c'}, {0x010C,'C'}, {0x010D,'c'}, {0x010F,'d'}, {0x011C,'G'},
	{0x011D,'g'}, {0x0124,'H'}, {0x0125,'h'}, {0x0132,''}, {0x0133,''},
	{0x0134,'J'}, {0x0135,'j'}, {0x013E,'l'}, {0x013F,'L'}, {0x0140,'l'},
	{0x0149,'\\'}, {0x0152,''}, {0x0153,''}, {0x015C,'S'}, {0x015D,'s'},
	{0x0160,'S'}, {0x0161,'s'}, {0x016D,'u'}, {0x0178,'Y'}, {0x017D,'Z'},
	{0x017E,'z'}, {0x017F,''}, {0x01C0,'|'}, {0x01C1,'|'}, {0x01C3,'!'},
	{0x01C7,'L'}, {0x01C8,'L'}, {0x01C9,'l'}, {0x01CA,'N'}, {0x01CB,'N'},
	{0x01CC,'n'}, {0x01DD, 26}, {0x0259, 26}, {0x02A3,'d'}, {0x02A6,'t'},
	{0x02AA,'l'}, {0x02AB,'l'}, {0x02B9,'\\'}, {0x02BA, 34}, {0x02BC,'\\'},
	{0x02C2,'<'}, {0x02C3,'>'}, {0x02C6,'^'}, {0x02C8,'\\'}, {0x02C9,''},
	{0x02CA,''}, {0x02CB,'`'}, {0x02CD,'_'}, {0x02D0,':'}, {0x02D1,''},
	{0x02D6,'+'}, {0x02D7,'-'}, {0x02DA,''}, {0x02DC,'~'}, {0x02EE, 34},
	{0x034F, 0 }, {0x0374,''}, {0x037E,';'}, {0x0391,'A'}, {0x0392,'B'},
	{0x0393,''}, {0x0394,127}, {0x0395,'E'}, {0x0396,'Z'}, {0x0397,'H'},
	{0x0398,''}, {0x0399,'I'}, {0x039A,'K'}, {0x039C,'M'}, {0x039D,'N'},
	{0x039E,''}, {0x039F,'O'}, {0x03A0,''}, {0x03A1,'P'}, {0x03A3,''},
	{0x03A4,'T'}, {0x03A5,'Y'}, {0x03A6,''}, {0x03A7,'X'}, {0x03A9,''},
	{0x03B1,''}, {0x03B2,''}, {0x03B4,''}, {0x03B5,''}, {0x03BC,''},
	{0x03BD,'v'}, {0x03BF,'o'}, {0x03C0,''}, {0x03C1,'p'}, {0x03C2,'s'},
	{0x03C3,''}, {0x03C4,''}, {0x03C6,''}, {0x03D5,''}, {0x03DC,'F'},
	{0x03F2,'c'}, {0x03F3,'j'}, {0x03F4,''}, {0x03F5,''}, {0x05BE,''},
	{0x05C0,'|'}, {0x05C3,':'}, {0x05D0,''}, {0x05D1,''}, {0x05D2,''},
	{0x05D3,''}, {0x05D4,''}, {0x05D5,''}, {0x05D6,''}, {0x05D7,''},
	{0x05D8,''}, {0x05D9,''}, {0x05DA,''}, {0x05DB,''}, {0x05DC,''},
	{0x05DD,''}, {0x05DE,''}, {0x05DF,''}, {0x05E0,''}, {0x05E1,''},
	{0x05E2,''}, {0x05E3,''}, {0x05E4,''}, {0x05E5,''}, {0x05E6,''},
	{0x05E7,''}, {0x05E8,''}, {0x05E9,''}, {0x05EA,''}, {0x05F0,''},
	{0x05F1,''}, {0x05F2,''}, {0x05F3,'\\'}, {0x05F4, 34}, {0x0660,'0'},
	{0x0661,'1'}, {0x0662,'2'}, {0x0663,'3'}, {0x0664,'4'}, {0x0665,'5'},
	{0x0666,'6'}, {0x0667,'7'}, {0x0668,'8'}, {0x0669,'9'}, {0x066A,'%'},
	{0x066B,','}, {0x066C,'\\'}, {0x066D,'*'}, {0x070F, 0 }, {0x09F3,'I'},
	{0x0E3F,'T'}, {0x10FB,':'}, {0x17DB,'K'}, {0x180B, 0 }, {0x180C, 0 },
	{0x180D, 0 }, {0x180E, 0 }, {0x2000,' '}, {0x2001,' '}, {0x2002,' '},
	{0x2003,' '}, {0x2004,' '}, {0x2005,' '}, {0x2006,' '}, {0x2007,' '},
	{0x2008,' '}, {0x2009,' '}, {0x200A, 0 }, {0x200B, 0 }, {0x200C, 0 },
	{0x200D, 0 }, {0x2010,'-'}, {0x2011,'-'}, {0x2012,'-'}, {0x2013,'-'},
	{0x2014,'-'}, {0x2015,'-'}, {0x2016,'|'}, {0x2018,'\\'}, {0x2019,'\\'},
	{0x201A,'\\'}, {0x201B,'\\'}, {0x201C, 34}, {0x201D, 34}, {0x201E, 34},
	{0x201F, 34}, {0x2020,''}, {0x2021,''}, {0x2022,''}, {0x2023,''},
	{0x2024,'.'}, {0x2025,'.'}, {0x2026,'.'}, {0x2027,''}, {0x202F,' '},
	{0x2030,''}, {0x2032,'\\'}, {0x2333, 34}, {0x2039,'<'}, {0x203A,'>'},
	{0x203C,'!'}, {0x203E,''}, {0x2043,'-'}, {0x2044,'/'}, {0x2047,'?'},
	{0x204A,'7'}, {0x204C,''}, {0x204D,''}, {0x204E,'*'}, {0x2052,'%'},
	{0x2057, 34}, {0x205F,' '}, {0x2060, 0 }, {0x2061, 0 }, {0x2062,' '},
	{0x2063,' '}, {0x207F,''}, {0x20A4,''}, {0x20A7,'P'}, {0x20A8,'R'},
	{0x20AA,'I'}, {0x20AC,''}, {0x20B0,'P'}, {0x2100,'a'}, {0x2101,'a'},
	{0x2103,''}, {0x2105,'c'}, {0x2106,'c'}, {0x2109,''}, {0x2113,'l'},
	{0x2114,'l'}, {0x2116,'N'}, {0x2122,''}, {0x2126,''}, {0x2127,'S'},
	{0x212A,'K'}, {0x212B,''}, {0x212E,'e'}, {0x2133,'M'}, {0x2135,''},
	{0x2136,''}, {0x2137,''}, {0x2138,''}, {0x2139,'i'}, {0x2180,''},
	{0x2190, 4 }, {0x2191, 1 }, {0x2192, 3 }, {0x2193, 2 }, {0x21E6, 4 },
	{0x21E7, 1 }, {0x21E8, 3 }, {0x21E9, 2 }, {0x2205,''}, {0x2206,127},
	{0x2208,''}, {0x220A,''}, {0x2211,''}, {0x2212,'-'}, {0x2215,'/'},
	{0x2216,'\\'}, {0x2217,'*'}, {0x2219,''}, {0x221A,''}, {0x221D,''},
	{0x221E,''}, {0x2223,'|'}, {0x2225,'|'}, {0x2227,''}, {0x2229,''},
	{0x222A,'U'}, {0x222E,''}, {0x2236,':'}, {0x223C,'~'}, {0x223F,'~'},
	{0x2248,''}, {0x2261,''}, {0x2264,''}, {0x2265,''}, {0x2295,'('},
	{0x2297,'('}, {0x2299,'('}, {0x22BA,'T'}, {0x22C0,''}, {0x22C2,''},
	{0x22C3,'U'}, {0x22C5,''}, {0x22EF,''}, {0x22FF,'E'}, {0x2310,''},
	{0x2314, 8 }, {0x231A, 9 }, {0x2320,''}, {0x2321,''}, {0x2329,'<'},
	{0x232A,'>'}, {0x23B7,''}, {0x240C, 12}, {0x240D, 13}, {0x241B, 27},
	{0x266A, 11}, {0x2680,'['}, {0x2681,'['}, {0x2682,'['}, {0x2683,'['},
	{0x2684,'['}, {0x2685,'['}, {0x2713, 8 }, {0x2714, 8 }, {0x27E8,'<'},
	{0x27E9,'>'}, {0x29F2,''}, {0x2A0D,''}, {0x2A2F,'x'}, {0x3000,' '}, 
	{0x301D, 34}, {0x301E, 34}, {0x301F, 34}, {0xF8F0,'j'}, {0xFB00,'f'},
	{0xFB01,'f'}, {0xFB02,'f'}, {0xFB03,'f'}, {0xFB04,'f'}, {0xFB05,''},
	{0xFB06,'s'}, {0xFB20,''}, {0xFB21,''}, {0xFB22,''}, {0xFB23,''},
	{0xFB24,''}, {0xFB25,''}, {0xFB26,''}, {0xFB27,''}, {0xFB28,''},
	{0xFB29,'+'}, {0xFB4F,''}, {0xFD3E,'('}, {0xFD3F,')'}, {0xFE00, 0 },
	{0xFE01, 0 }, {0xFE02, 0 }, {0xFE03, 0 }, {0xFE04, 0 }, {0xFE05, 0 },
	{0xFE06, 0 }, {0xFE07, 0 }, {0xFE08, 0 }, {0xFE09, 0 }, {0xFE0A, 0 },
	{0xFE0B, 0 }, {0xFE0C, 0 }, {0xFE0D, 0 }, {0xFE0E, 0 }, {0xFE0F, 0 },
	{0xFEFF, 0 },
	{0xFFFD, 6 } /* default value */
};

/*--------------------------- ROUTINES ---------------------------------------*/

/* cookies */

static COOKIE *fcookie(void)
{
	COOKIE *p;
	long stack;
	stack = 0;
	if(Super(1L) >= 0)
		stack = Super(0L);
	p = *(COOKIE **)_p_cookies;
	if(stack)
		Super((void *)stack);
	if(!p)
		return((COOKIE *)0);
	return(p);
}

static COOKIE *ncookie(COOKIE *p)
{
	if(!p->ident)
		return(0);
	return(++p);
}

static COOKIE *get_cookie(long id)
{
	COOKIE *p;
	p = fcookie();
	while(p)
	{
		if(p->ident == id)
			return p;
		p = ncookie(p);
	}
	return((COOKIE *)0);
}

/* Drag & Drop */

static char pipename[] = "U:\\PIPE\\DRAGDROP.AA";
static long pipesig;

static short ddgetsig(long *oldsig)
{
	*oldsig = (long)Psignal(SIGPIPE, (void *)SIG_IGN);
}

static void ddsetsig(long oldsig)
{
	if(oldsig != -32L)
		Psignal(SIGPIPE, (void *)oldsig);
}

static void ddclose(short fd)
{
	ddsetsig(pipesig); /* restore Signalhandler */
	Fclose(fd);
}

static short ddopen(short ddnam, char ddmsg)
{
	long fd;
	pipename[17] = (ddnam & 0xff00) >> 8;
	pipename[18] = ddnam & 0x00ff;
	fd = Fopen(pipename, 2);
	if(fd < 0L)
		return(-1);
	ddgetsig(&pipesig); 	/* Signalhandler config */
	if(Fwrite((short)fd, 1L, &ddmsg) != 1L)
	{
		ddclose((short)fd);
		return(-1);
	}
	return((short)fd);
}

static short ddsexts(short fd, char *exts)
{
	if(Fwrite(fd, DD_EXTSIZE, exts) != DD_EXTSIZE)
		return(-1);
	return(1);
}

static short ddreply(short fd, char ack)
{
	if(Fwrite(fd, 1L, &ack) != 1L)
		return(-1);
	return(1);
}

/*
*	Read inside the pipe who is normally:
*	1 short: header size
*	4 bytes: data type (whichext)
*	1 long:	data(s) size
*	string: data(s) descriptor
*	string: name of file(s)
*	=> 11 bytes min, max string size 128 bytes
*/
static short ddrtry(short fd, char *name, char *file, char *whichext, long *size)
{
	char buf[DD_NAMEMAX * 2];
	short hdrlen, i, len;
	if(Fread(fd, 2L, &hdrlen) != 2L)
		return(-1);
	if(hdrlen < 9)                    /* minimum of 11 - 2 = 9 bytes to read */
		return(-1);                     /* bad size */
	if(Fread(fd, 4L, whichext) != 4L) /* read ext */
		return(-1);	
	if(Fread(fd, 4L, size) != 4L)     /* read datas size */
		return(-1);	
	hdrlen -= 8;	/* we have read 8 bytes */
	if(hdrlen > DD_NAMEMAX*2)
		i = DD_NAMEMAX*2;
	else
		i = hdrlen;
	len = i;
	if(Fread(fd, (long)i, buf) != (long)i)
		return(-1);
	hdrlen -= i;
	strncpy(name, buf, DD_NAMEMAX);
	i = (short)strlen(name) + 1;
	if(len - i > 0)
		strncpy(file, buf + i, DD_NAMEMAX);
	else
		file[0] = '\0';
	while(hdrlen > DD_NAMEMAX*2)
	{
		if(Fread(fd, DD_NAMEMAX*2, buf) != DD_NAMEMAX*2)
			return(-1);	
		hdrlen -= DD_NAMEMAX*2;
	}
	if(hdrlen > 0)
	{
		if(Fread(fd, (long)hdrlen, buf) != (long)hdrlen)
			return(-1);
	}
	return(1);
}

static void gem_get_size(struct graphics_device *dev)
{
	struct gem_window *view;
	if((dev == NULL) || (dev->driver_data == NULL)) 
		return;	
	view = (struct gem_window *)dev->driver_data;
	if((view->handle >= 0) && !view->iconified)
	{
		wind_get(view->handle, WF_WORKXYWH, &view->x, &view->y, &view->w, &view->h);
		view->buttons = 0;
		if(buttons != NULL)
		{
			buttons[BBOX].ob_x = view->x;
			buttons[BBOX].ob_y = view->y;
			buttons[BBOX].ob_width = view->w;
			if(view->h > buttons[BBOX].ob_height)
			{
				view->y += buttons[BBOX].ob_height;
				view->h -= buttons[BBOX].ob_height;
				view->buttons = 1;
			}
		}
		dev->size.x1 = dev->size.y1 = 0;
		dev->size.x2 = (int)view->w + 1;
		dev->size.y2 = (int)view->h + 1;
	}
}

/* text convert routines */

static int atari_to_iso_8859_1(int c)
{
	switch(c)
	{
		case 0xAD: c = 0xA1; break; // iexcl
		case 0x9B: c = 0xA2; break; // cent
		case 0x9C: c = 0xA3; break; // pound
		case 0x9D: c = 0xA5; break; // yen
		case 0xDD: c = 0xA7; break; // sect
		case 0xB9: c = 0xA8; break; // uml
		case 0xBD: c = 0xA9; break; // reg
		case 0xAE: c = 0xAB; break; // laquo
		case 0xAA: c = 0xAC; break; // not
		case 0xBE: c = 0xAE; break; // copy
		case 0xF8: c = 0xB0; break; // deg
		case 0xF1: c = 0xB1; break; // plusmn
		case 0xFD: c = 0xB2; break; // sup2
		case 0xFE: c = 0xB3; break; // sup3
		case 0xBA: c = 0xB4; break; // acute
		case 0xE6: c = 0xB5; break; // micro
		case 0xBC: c = 0xB6; break; // para
		case 0xAF: c = 0xBB; break; // raquo
		case 0xAC: c = 0xBC; break; // frac14
		case 0xAB: c = 0xBD; break; // frac12
		case 0xA8: c = 0xBF; break; // iquest
		case 0xB6: c = 0xC0; break; // Agrave
		case 0xB7: c = 0xC3; break; // Atilde
		case 0x8E: c = 0xC4; break; // Auml
		case 0x8F: c = 0xC5; break; // Aring
		case 0x92: c = 0xC6; break; // Aelig
		case 0x80: c = 0xC7; break; // Ccedil
		case 0x90: c = 0xC9; break; // Eacute
		case 0xB8: c = 0xD5; break; // Otilde
		case 0xB2: c = 0xD8; break; // Oslash
		case 0x9A: c = 0xDC; break; // Uuml
		case 0xE1: c = 0xDF; break; // szlig
	  case 0x85: c = 0xE0; break; // agrave
		case 0xA0: c = 0xE1; break; // aacute
		case 0x83: c = 0xE2; break; // acirc
		case 0xB0: c = 0xE3; break; // atilde
		case 0x84: c = 0xE4; break; // auml
		case 0x86: c = 0xE5; break; // aring
		case 0x91: c = 0xE6; break; // aelig
		case 0x87: c = 0xE7; break; // ccedil
		case 0x8A: c = 0xE8; break; // egrave
		case 0x82: c = 0xE9; break; // eacute
		case 0x88: c = 0xEA; break; // ecirc
		case 0x89: c = 0xEB; break; // euml
		case 0x8D: c = 0xEC; break; // igrave
		case 0xA1: c = 0xED; break; // iacute
		case 0x8C: c = 0xEE; break; // icirc
		case 0x8B: c = 0xEF; break; // iuml
		case 0xA4: c = 0xF1; break; // ntilde
		case 0x95: c = 0xF2; break; // ograve
		case 0xA2: c = 0xF3; break; // oacute		
		case 0x93: c = 0xF4; break; // ocirc
		case 0xB1: c = 0xF5; break; // otilde
		case 0x94: c = 0xF6; break; // ouml
		case 0xF6: c = 0xF7; break; // divide
		case 0xB3: c = 0xF8; break; // oslash
		case 0x97: c = 0xF9; break; // ugrave
		case 0xA3: c = 0xFA; break; // uacute
		case 0x96: c = 0xFB; break; // ucirc
		case 0x81: c = 0xFC; break; // uuml
		case 0x98: c = 0xFF; break; // yuml
		default:
			if((c < ' ') || (c >= 127))
				c = ' ';
			break;
	}
	return(c);
}

#if 0 /* slower */

static int iso_8859_1_to_atari(int c)
{
	switch(c)
	{
		case 0xA1: c = 0xAD; break; // iexcl
		case 0xA2: c = 0x9B; break; // cent
		case 0xA3: c = 0x9C; break; // pound
		case 0xA5: c = 0x9D; break; // yen
		case 0xA7: c = 0xDD; break; // sect
		case 0xA8: c = 0xB9; break; // uml
		case 0xA9: c = 0xBD; break; // reg
		case 0xAB: c = 0xAE; break; // laquo
		case 0xAC: c = 0xAA; break; // not
		case 0xAE: c = 0xBE; break; // copy
		case 0xB0: c = 0xF8; break; // deg
		case 0xB1: c = 0xF1; break; // plusmn
		case 0xB2: c = 0xFD; break; // sup2
		case 0xB3: c = 0xFE; break; // sup3
		case 0xB4: c = 0xBA; break; // acute
		case 0xB5: c = 0xE6; break; // micro
		case 0xB6: c = 0xBC; break; // para
		case 0xBB: c = 0xAF; break; // raquo
		case 0xBC: c = 0xAC; break; // frac14
		case 0xBD: c = 0xAB; break; // frac12
		case 0xBF: c = 0xA8; break; // iquest
		case 0xC0: c = 0xB6; break; // Agrave
		case 0xC3: c = 0xB7; break; // Atilde
		case 0xC4: c = 0x8E; break; // Auml
		case 0xC5: c = 0x8F; break; // Aring
		case 0xC6: c = 0x92; break; // Aelig
		case 0xC7: c = 0x80; break; // Ccedil
		case 0xC9: c = 0x90; break; // Eacute
		case 0xD5: c = 0xB8; break; // Otilde
		case 0xD8: c = 0xB2; break; // Oslash
		case 0xDC: c = 0x9A; break; // Uuml
		case 0xDF: c = 0xE1; break; // szlig
	  case 0xE0: c = 0x85; break; // agrave
	  case 0xE1: c = 0xA0; break; // aacute
		case 0xE2: c = 0x83; break; // acirc
		case 0xE3: c = 0xB0; break; // atilde
		case 0xE4: c = 0x84; break; // auml
		case 0xE5: c = 0x86; break; // aring
		case 0xE6: c = 0x91; break; // aelig
		case 0xE7: c = 0x87; break; // ccedil
		case 0xE8: c = 0x8A; break; // egrave
		case 0xE9: c = 0x82; break; // eacute
		case 0xEA: c = 0x88; break; // ecirc
		case 0xEB: c = 0x89; break; // euml
		case 0xEC: c = 0x8D; break; // igrave
		case 0xED: c = 0xA1; break; // iacute
		case 0xEE: c = 0x8C; break; // icirc
		case 0xEF: c = 0x8B; break; // iuml
		case 0xF1: c = 0xA4; break; // ntilde
		case 0xF2: c = 0x95; break; // ograve
		case 0xF3: c = 0xA2; break; // oacute
		case 0xF4: c = 0x93; break; // ocirc
		case 0xF5: c = 0xB1; break; // otilde
		case 0xF6: c = 0x94; break; // ouml
		case 0xF7: c = 0xF6; break; // divide
		case 0xF8: c = 0xB3; break; // oslash
		case 0xF9: c = 0x97; break; // ugrave
		case 0xFA: c = 0xA3; break; // uacute
		case 0xFB: c = 0x96; break; // ucirc
		case 0xFC: c = 0x81; break; // uuml
		case 0xFF: c = 0x98; break; // yuml
		default:
			if((c < ' ') || (c >= 127))
				c = ' ';
			break;
	}
	return(c);
}

static int gem_convert_text(unsigned char *text)
{
	int c, len = 0;
	struct conv_table *ct;
	unsigned char *p, *p2;
  ct = get_translation_table(utf8_table, get_cp_index("ISO-8859-1"));
	p = p2 = convert_string(ct, text, strlen(text), NULL);
	while((c = *p2 & 0xFF) != 0)
	{
		c= iso_8859_1_to_atari(c);
		*p2++ = (unsigned char)c;
		len++;
	}
	strcpy(text, p);
	mem_free(p);
	return(len);
}

#else /* faster */

static unsigned char utf8_to_atari(unsigned char **text)
{
	unsigned char *s = *text;
	unsigned long unicode = (unsigned long)*(s++);
	int utf8ok = 1, bytes = 0 , mask, shift, i;
	const struct s_uni_trans *tab = Unicode_to_Atari;
	int end = (sizeof(Unicode_to_Atari) / sizeof(*Unicode_to_Atari)) - 1;
	int begin = 0;
	unsigned char code = tab[end].atari; /* default value */
	if((unicode >= 0xC0) && (unicode <= 0xFD) && (s[0] >= 0x80) && (s[0] <= 0xBF))
	{
		bytes++;  /* 1 byte for U+0080..U+07FF */
		if(unicode > 0xDF)
		{
			if((s[1] >= 0x80) && (s[1] <= 0xBF))
			{
				bytes++;  /* 2 bytes for U+0800..U+FFFF */
				if(unicode > 0xEF)
				{
					if((s[2] >= 0x80) && (s[2] <= 0xBF))
					{
						bytes++;  /* 3 bytes for U+10000..U+1FFFFF */
						if(unicode > 0xF7)
						{
							if((s[3] >= 0x80) && (s[3] <= 0xBF)) {
								bytes++;  /* 4 bytes for U+200000..U+3FFFFFF */
								if(unicode > 0xFB)
								{
									if((s[4] >= 0x80) && (s[4] <= 0xBF))
										bytes++;  /* 5 bytes for U+4000000..U+7FFFFFFF */
									else
										utf8ok = 0;
								}
							}
							else
								utf8ok = 0;
						}
					}
					else
						utf8ok = 0;
				}
			}
			else
				utf8ok = 0;
		}
	}
	else
		utf8ok = 0;
	/* nonshortest forms are illegal */
	if(!utf8ok || (unicode < 0xC2) || ((unicode == 0xE0) && (s[0] < 0xA0))
	 || ((unicode == 0xF0) && (s[0] < 0x90)) || ((unicode == 0xF8) && (s[0] < 0x88))
	 || ((unicode == 0xFC) && (s[0] < 0x84)))
	{
		unicode = (unsigned long)tab[end].unicode;  /* default value */
		s += bytes;
	}
	else 
	{
		/* calculate value */
		mask = (1 << (6 - bytes)) - 1;  /* 0x1F, 0x0F, 0x07, 0x03, 0x01 */
		shift = bytes * 6;              /* 30, 24, 18, 12, 6 */
		unicode = (unicode & mask) << shift;
		for(i = 0, shift -= 6; i < bytes ; i++, shift -= 6)
			unicode += (unsigned long)(*s++ & 0x3F) << shift;
	}
	/* unicodes not supported */
	if((unicode >= 0xD800 && unicode <= 0xDFFF) || unicode >= 0xFFFE)
		unicode = (unsigned long)tab[end].unicode;  /* default value */
	do
	{
		int i = (end + begin) / 2;
		if(unicode > (unsigned long)tab[i].unicode)
			begin = i + 1;
		else if(unicode < (unsigned long)tab[i].unicode)
			end = i - 1;
		else
		{
			code = tab[i].atari;
			break;
		}
	}
	while(begin <= end);
	*text = s; /* update source pointer */
	return(code);
}

static int gem_convert_text(unsigned char *text)
{
	int len = 0;
	unsigned char *p, *p2;
	p = p2 = text;
	while(*p)
	{
		if(*p < ' ')
		{
			p++;
			continue;
		}
		else if(*p < 0x80)
			*p2++ = *p++;
		else
			*p2++ = utf8_to_atari(&p);
		len++;
	}
	*p2 = '\0';
	return(len);
}

#endif

/* gem events */

static void gem_send_redraw(struct graphics_device *dev)
{
	short msg[8];
	struct gem_window *view;
	if((dev == NULL) || (dev->driver_data == NULL)) 
		return;	
	view = (struct gem_window *)dev->driver_data;
	if(view->handle >= 0)
	{
		msg[0] = WM_REDRAW;
		msg[1] = ap_id;
		msg[2] = 0;
		msg[3] = view->handle;
		if(view->buttons)
		{
			msg[4] = view->x;
			msg[5] = view->y - buttons[BBOX].ob_height;
			msg[6] = view->w;
			msg[7] = view->h + buttons[BBOX].ob_height;
		}
		else
		{
			msg[4] = view->x;
			msg[5] = view->y;
			msg[6] = view->w;
			msg[7] = view->h;
		}
		appl_write(ap_id, 16, msg);
	}
}

static int gem_test_win_top(struct graphics_device *dev, int x1, int y1, int x2, int y2)
{
	short msg[8];
	struct gem_window *view;
	if((dev == NULL) || (dev->driver_data == NULL)) 
		return;	
	view = (struct gem_window *)dev->driver_data;
	if(view->handle >= 0)
	{
		short win_top, dummy;
		if(view->flag_redraw)
			return(1);
		wind_get(view->handle, WF_TOP, &win_top, &dummy, &dummy, &dummy);
		if(view->handle == win_top)
			return(1);
		msg[0] = WM_REDRAW;
		msg[1] = ap_id;
		msg[2] = 0;
		msg[3] = view->handle;
#if 0 /* slower */
		msg[4] = (short)x1 + view->x;
		msg[5] = (short)y1 + view->y;
		msg[6] = (short)x2 + view->x - 1;
		msg[7] = (short)y2 + view->y - 1;
		msg[6] -= (msg[4] - 1); /* x, y, w, h */
		msg[7] -= (msg[5] - 1);
#else /* faster */
		msg[4] = view->x;
		msg[5] = view->y;
		msg[6] = view->w;
		msg[7] = view->h;
#endif
		appl_write(ap_id, 16, msg);
	}
	return(0);
}

static void gem_open_files(struct graphics_device *dev, char *files)
{
  static char file[] = "file://";
	int i = 0, spaces = 0;
	char c;
	dev->keyboard_handler(dev, 'g', 0); /* URL */
	while(file[i])
		dev->keyboard_handler(dev, file[i++], 0);
	i = 0;
	if(files[0] == 0x27)
	{
		spaces = 1;
		i++;
	}		
	while((c = files[i++]) != 0)
	{
		if((spaces && (c == 0x27))
		 || (!spaces && (c == ' '))) /* only 1st file */
			break;
		if(c == '\\')
			c = '/';
		dev->keyboard_handler(dev, c, 0);
	}
	dev->keyboard_handler(dev, KBD_ENTER, 0);
}

/* This is a nasty hack */
#undef select

int gem_select(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout)
{
#define DELAY_LOOP_MS 10000
	struct timeval time_out;
	int retval, count, ev;
	short msg[8], rect[4], kc, kstate, prevkc, prevks;
	short mousex, mousey, mouseb, mclicks, win_top, dummy;
	static short last_x = -1, last_y = -1, last_b = -1;
	short handle, x, y, w, h;
	long timeout_ms;
	struct gem_window *view;
	if(drv != &gem_driver)
		return(select(n, readfds, writefds, exceptfds, timeout));
	time_out.tv_sec = 0;
	time_out.tv_usec = DELAY_LOOP_MS; /* select */
	if(timeout != NULL)
	  timeout_ms = (timeout->tv_usec/1000) + (timeout->tv_sec * 1000);
	else
		timeout_ms = 0x7FFFFFFF;
#ifdef DEBUG_GEM
	if(timeout != NULL)
		printf("gem_select n: %d, timeout: %d.%03d S", n, timeout->tv_sec, timeout->tv_usec/1000);
	else
		printf("gem_select n: %d, infinite timeout", n);
	printf(", %s%s%s\r\n", readfds ? "read " : "", writefds ? "write " : "", exceptfds ? "except" : "");
#endif
	wind_update(END_UPDATE);
	ev = 0;
	do
	{
		retval = (int)evnt_multi(MU_MESAG | MU_TIMER |  MU_BUTTON | MU_KEYBD,
		 258, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		 msg, DELAY_LOOP_MS/1000, &mousex, &mousey, &mouseb, &kstate, &kc, &mclicks);
		/* Message event ? */
		if(retval & MU_MESAG)
		{
			for(count = 0; count < MAX_DEV; count++)
			{
				struct graphics_device *dev = gem_dev[count];
				if((dev == NULL) || !vdi_handle)
					continue;
				view = (struct gem_window *)dev->driver_data;
				if((view == NULL) || (view->handle < 0) || (view->handle != msg[3]))
					continue;
				switch(msg[0])
				{
					case WM_CLOSED:
					case AP_TERM:    
						dev->keyboard_handler(dev, KBD_CTRL_C, 0); /* quit */
						ev = 1;
						break;
					case WM_MOVED:
						wind_set(msg[3], WF_CURRXYWH, msg[4], msg[5], msg[6], msg[7]);
						gem_get_size(dev);
						break;
					case WM_TOPPED:
						wind_set(msg[3], WF_TOP, 0, 0, 0, 0);
						break;
					case WM_REDRAW:
						wind_update(BEG_UPDATE);
						/* Browse the rectangle list to redraw */
						if(wind_get(msg[3], WF_FIRSTXYWH, &rect[0], &rect[1], &rect[2], &rect[3])
						 && rect[2] && rect[3])
						{
							int hide = 0;
							do
							{
								if(rc_intersect((GRECT *)&msg[4], (GRECT *)rect))
								{
									if(!hide)
									{
										v_hide_c(vdi_handle);
										hide = 1;
									}
									if(view->iconified)
									{
										short icon_rect[4];
										/* Clear the rectangle */
										vsf_color(vdi_handle, 0);
										vsf_interior(vdi_handle, 1);
										vsf_perimeter(vdi_handle, 0);
										vswr_mode(vdi_handle, MD_REPLACE);
										rect[2] += (rect[0] - 1);
										rect[3] += (rect[1] - 1);
										vs_clip(vdi_handle, 1, rect);
										v_bar(vdi_handle, rect);
										rect[2] -= (rect[0] - 1); /* x, y, w, h */
										rect[3] -= (rect[1] - 1);
                    /* Put icon to center of the reduced window */
										icon_rect[0] = view->x + ((view->w - (short)links_image.width) >> 1);
										icon_rect[1] = view->y + ((view->h - (short)links_image.height) >> 1);
										icon_rect[2] = (short)links_image.width;
										icon_rect[3] = (short)links_image.height;
										if(rc_intersect((GRECT *)icon_rect, (GRECT *)rect))
										{
											MFDB mfdb_src, mfdb_dst;
											short pxy[8];
											int bmp_size;
											void *buf = NULL;
											rect[2] += (rect[0] - 1);
											rect[3] += (rect[1] - 1);
											vs_clip(vdi_handle, 1, rect);
											pxy[0] = pxy[1] = 0;
											pxy[2] = (short)links_image.width - 1;
											pxy[3] = (short)links_image.height - 1;
											pxy[4] = icon_rect[0];
											pxy[5] = icon_rect[1];
											pxy[6] = icon_rect[0] + icon_rect[2]  - 1;
											pxy[7] = icon_rect[1] + icon_rect[3]  - 1;
											mfdb_src.fd_addr = icon;
											mfdb_src.fd_w = (short)((links_image.width + 15) & ~15); /* alignment */
											mfdb_src.fd_h = (short)links_image.height;
											mfdb_src.fd_nplanes = (short)bpp;
											mfdb_src.fd_wdwidth = mfdb_src.fd_w >> 4;
											mfdb_src.fd_stand = mfdb_src.fd_r1 = mfdb_src.fd_r2 = mfdb_src.fd_r3 = 0;
											memset(&mfdb_dst, 0, sizeof(MFDB));
											bmp_size = (((int)links_image.width + 15) & ~15) * (int)links_image.height;
											if((pixelsize == 1) && !packed_format
											 && ((buf = mem_alloc(bmp_size)) != NULL))
											{
												if(bpp == 4)
													Atari_C2pConvert4(icon, buf, (unsigned long)((links_image.width + 15) & ~15), (unsigned long)links_image.height, 0, (unsigned long)links_image.width, (((unsigned long)links_image.width + 15) & ~15) >> 1);
												else
													Atari_C2pConvert8(icon, buf, (unsigned long)((links_image.width + 7) & ~7), (unsigned long)links_image.height, 0, (unsigned long)links_image.width, ((unsigned long)links_image.width + 15) & ~15);
												mfdb_src.fd_addr = buf;
											}
											vro_cpyfm(vdi_handle, S_ONLY, pxy, &mfdb_src, &mfdb_dst);
											if(buf != NULL)
												mem_free(buf);
                    }
									}
									else
									{
										if((buttons != NULL) && view->buttons)
										{
											buttons[BBOX].ob_x = view->x;
											buttons[BBOX].ob_y = view->y - buttons[BBOX].ob_height;
											objc_draw(buttons, BBOX, MAX_DEPTH, rect[0], rect[1], rect[2], rect[3]);
										}
										view->redraw.x = rect[0];
										view->redraw.y = rect[1];
										view->redraw.w = rect[2];
										view->redraw.h = rect[3];
										rect[2] += (rect[0] - 1);
										rect[3] += (rect[1] - 1);
										vs_clip(vdi_handle, 1, rect);
										view->flag_redraw = 1;
										dev->resize_handler(dev);
										view->flag_redraw = 0;
									}
								}
							}
							while(wind_get(msg[3], WF_NEXTXYWH, &rect[0], &rect[1], &rect[2], &rect[3])
							 && rect[2] && rect[3]);
							if(hide)
								v_show_c(vdi_handle, 1);
						}
						wind_update(END_UPDATE);
						break;
					case WM_ICONIFY:
					case WM_ALLICONIFY:
						wind_update(BEG_UPDATE);
						graf_shrinkbox(msg[4], msg[5], msg[6], msg[7], view->x, view->y, view->w, view->h);
						wind_set(msg[3], WF_ICONIFY, msg[4], msg[5], msg[6], msg[7]);
						view->iconified = 1;
						wind_get(msg[3], WF_WORKXYWH, &view->x, &view->y, &view->w, &view->h);
						wind_update(END_UPDATE);
						gem_send_redraw(dev);
						break;
					case WM_UNICONIFY:
						wind_update(BEG_UPDATE);
						graf_growbox(view->x, view->y, view->w, view->h, msg[4], msg[5], msg[6], msg[7]);
						wind_set(msg[3], WF_UNICONIFY, msg[4], msg[5], msg[6], msg[7]);
						view->iconified = 0;
						wind_update(END_UPDATE);
						gem_get_size(dev);
						gem_send_redraw(dev);
						break;
					case WM_SIZED:
						wind_set(msg[3], WF_CURRXYWH, msg[4], msg[5], msg[6], msg[7]);
						wind_calc(1, view->type, msg[4], msg[5], msg[6], msg[7], &x, &y, &w, &h);
						gem_get_size(dev);
						gem_send_redraw(dev);
						view->fulled = 0;
						break;
					case WM_FULLED:
						if(view->fulled)
						{
							wind_get(msg[3], WF_PREVXYWH, &x, &y, &w, &h);
							view->fulled = 0;
						}
						else
						{
							x = desk_x;
							y = desk_y;
							w = desk_w;
							h = desk_h;
							view->fulled = 1;
						}
						wind_set(msg[3], WF_CURRXYWH, x, y, w, h);
						wind_calc(1, view->type, x, y, w, h, &x, &y, &w, &h);
						gem_get_size(dev);
						gem_send_redraw(dev);
						break;
					case WM_BOTTOMED:
					case WM_UNTOPPED:
						break;
					case AP_DRAGDROP:				
						// msg[3]=mx, msg[4]=my, msg[6]=kstate
						if((handle = ddopen(msg[7], DD_OK)) >= 0)
						{
							static char exts[DD_EXTSIZE] = "ARGS.TXT\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
							if(ddsexts(handle, exts) >= 0)
							{
								char name[DD_NAMEMAX], file[DD_NAMEMAX], whichext[4];
								long size;
								char *buf;
								while(ddrtry(handle, name, file, whichext, &size) >= 0)
								{
									if(!(whichext[0] == 'A' && whichext[1] == 'R' && whichext[2] == 'G' && whichext[3] == 'S'))
									{
										if(ddreply(handle, DD_EXT) < 0)
											break;
										else
											continue;
									}
									if((buf = mem_alloc(size + 1)) == NULL)
									{
										if(ddreply(handle, DD_LEN) < 0)
											break;
										else
											continue;
									}
									else
									{
										if((ddreply(handle, DD_OK) >= 0)
										 && (Fread(handle, size, buf) == size))
										{
											buf[size] = '\0';
											gem_open_files(dev, buf);
											ev = 1;
										}
										mem_free(buf);
										break;
									}
								}
							}
							ddclose(handle);	
						}
						break;
				}
				break;
			}
			if(msg[0] == VA_START)
			{
				char *p = (char *)*((long *)&msg[3]);
				if((p != NULL) && *p)
				{					
					short msg2[8];
					msg2[0] = AV_STARTED;
					msg2[1] = ap_id;
					msg2[2] = msg2[3] = msg2[4] = msg2[5] = msg2[6] = msg2[7] = 0;
					appl_write(msg[1], 16, msg2);
					wind_get(DESKTOP_HANDLE, WF_TOP, &win_top, &dummy, &dummy, &dummy);
					for(count = 0; count < MAX_DEV; count++)
					{
						struct graphics_device *dev = gem_dev[count];
						if((dev == NULL) || !vdi_handle)
							continue;
						view = (struct gem_window *)dev->driver_data;
						if((view == NULL) || (view->handle < 0) || (view->handle != win_top))
							continue;
						gem_open_files(dev, p);
						ev = 1;
						break;
					}
					if(!ev) /* window not found at top */
					{
						for(count = 0; count < MAX_DEV; count++)
						{
							struct graphics_device *dev = gem_dev[count];
							if((dev == NULL) || !vdi_handle)
								continue;
							view = (struct gem_window *)dev->driver_data;
							if((view == NULL) || (view->handle < 0))
								continue;
							wind_set(view->handle, WF_TOP, 0, 0, 0, 0);
							gem_open_files(dev, p);
							ev = 1;
							break;
						}
					}
				}
			}
		}
		/* Keyboard event ? */
		if(retval & MU_KEYBD)
		{
			if((prevkc != kc) || (prevks != kstate))
			{
				int c;
				switch(kc >> 8) /* scan-code */
				{
					case 0x3B: c = KBD_F1; break;
					case 0x3C: c = KBD_F2; break;
					case 0x3D: c = KBD_F3; break;
					case 0x3E: c = KBD_F4; break;
					case 0x3F: c = KBD_F5; break;
					case 0x40: c = KBD_F6; break;					
					case 0x41: c = KBD_F7; break;
					case 0x42: c = KBD_F8; break;
					case 0x43: c = KBD_F9; break;
					case 0x44: c = KBD_F10; break;
					case 0x4B: c = KBD_LEFT; break;
					case 0x4D: c = KBD_RIGHT; break;
					case 0x48:
						if(kstate & (K_RSHIFT|K_LSHIFT))
							c = KBD_PAGE_UP;
						else
							c = KBD_UP;
						break;
					case 0x50:
						if(kstate & (K_RSHIFT|K_LSHIFT))
							c = KBD_PAGE_DOWN;
						else
					 		c = KBD_DOWN;
					 	break;
					case 0x52: c = KBD_INS; break;
					case 0x47:
						if(kstate & (K_RSHIFT|K_LSHIFT))
							c = KBD_END;
						else
							c = KBD_HOME;
						break;
					default:
						switch (kc & 255) /* ascii */
						{
							case 8: c = KBD_BS; break;
							case 9: c = KBD_TAB; break;
							case 13: c = KBD_ENTER; break;
							case 27: c = KBD_ESC; break;
							case 127: c = KBD_DEL; break;
							default:
								c = (int)(kc & 255);
								if(c < 32)
								{
									if(c && (kstate & K_CTRL))
										c += 0x40;
									else
										c = 0;
								}
								else
									c = cp2u((unsigned char)atari_to_iso_8859_1(c), get_cp_index("ISO-8859-1"));
								break;
						}
						break;
				}
				if(c)
				{
					wind_get(DESKTOP_HANDLE, WF_TOP, &win_top, &dummy, &dummy, &dummy);
					for(count = 0; count < MAX_DEV; count++)
					{
						struct graphics_device *dev = gem_dev[count];
						if(dev == NULL)
							continue;
						view = (struct gem_window *)dev->driver_data;
						if((view == NULL) || (view->handle < 0)
						 || (view->handle != win_top) || view->iconified)
							continue;
						dev->keyboard_handler(dev, c, (kstate & (K_RSHIFT|K_LSHIFT) ? KBD_SHIFT : 0) | (kstate & K_CTRL ? KBD_CTRL : 0) | (kstate & K_ALT ? KBD_ALT : 0));
						ev = 1;
						break;
					}
				}
			}
		}
		/* Button event ? */
		if(retval & MU_BUTTON)
		{
			mouseb &= 3;
			if(mouseb == 3)
				mouseb++;
			if(mouseb != last_b)
			{
				short index;
				wind_get(DESKTOP_HANDLE, WF_TOP, &win_top, &dummy, &dummy, &dummy);
				for(count = 0; count < MAX_DEV; count++)
				{
					struct graphics_device *dev = gem_dev[count];
					if(dev == NULL)
						continue;
					view = (struct gem_window *)dev->driver_data;
					if((view == NULL) || (view->handle < 0)
					 || view->iconified)
						continue;
					if((buttons != NULL) && view->buttons && (mouseb == 1))
					{
						buttons[BBOX].ob_x = view->x;
						buttons[BBOX].ob_y = view->y - buttons[BBOX].ob_height;					
						switch(objc_find(buttons, BBOX, MAX_DEPTH, mousex, mousey))
						{
							case BBACK:
								dev->keyboard_handler(dev, KBD_LEFT, 0);
								ev = 1;
								break;
							case BNEXT:
								dev->keyboard_handler(dev, KBD_RIGHT, 0);
								ev = 1;
								break;
							case BSTOP:
								dev->keyboard_handler(dev, 'S', KBD_CTRL);
								ev = 1;
								break;
							case BRELOAD:
								dev->keyboard_handler(dev, 'R', KBD_CTRL);
								ev = 1;
								break;
							case BURL:
								dev->keyboard_handler(dev, 'g', 0);
								ev = 1;
								break;
							case BLIST:
								dev->keyboard_handler(dev, 's', 0);
								ev = 1;
								break;
						}
					}
					if(view->handle != win_top)
						continue;
					if((mousex >= view->x) && (mousey >= view->y)
					 && (mousex < view->x + view->w) && (mousey < view->y + view->h))
					{
						if(!mouseb)
						{
							dev->mouse_handler(dev, (int)(mousex - view->x), (int)(mousey - view->y),
							 B_UP | (last_b & 1 ? B_LEFT : last_b & 2 ? B_RIGHT : last_b & 4 ? B_MIDDLE : 0));
							ev = 1;
						}
						else
						{
							dev->mouse_handler(dev, (int)(mousex - view->x), (int)(mousey - view->y),
							 B_DOWN | (mouseb & 1 ? B_LEFT : mouseb & 2 ? B_RIGHT : mouseb & 4 ? B_MIDDLE : 0));
							ev = 1;
						}
					}
					break;
				}
				last_b = mouseb;
			}
		}
		/* Update mouse x, y */
		wind_get(DESKTOP_HANDLE, WF_TOP, &win_top, &dummy, &dummy, &dummy);
		graf_mkstate(&mousex, &mousey, &mouseb, &kstate);
		if(mouseb == 3)
			mouseb++;
		if((mousex != last_x) || (mousey != last_y) || (mouseb != last_b))
		{						
			for(count = 0; count < MAX_DEV; count++)
			{
				struct graphics_device *dev = gem_dev[count];
				if(dev == NULL)
					continue;
				view = (struct gem_window *)dev->driver_data;
				if((view == NULL) || (view->handle < 0)
				 || (view->handle != win_top) || view->iconified)
					continue;
				if((mousex >= view->x) && (mousey >= view->y)
				 && (mousex < view->x + view->w) && (mousey < view->y + view->h))
				{
					if((mouseb != last_b) && !mouseb)
					{
						dev->mouse_handler(dev, (int)(mousex - view->x), (int)(mousey - view->y),
						 B_UP | (last_b & 1 ? B_LEFT : last_b & 2 ? B_RIGHT : last_b & 4 ? B_MIDDLE : 0));
						ev = 1;
					}
					if((mousex != last_x) && (mousey != last_y))
					{
						dev->mouse_handler(dev, (int)(mousex - view->x), (int)(mousey - view->y),
						 !mouseb ? B_MOVE : B_DRAG | (mouseb & 1 ? B_LEFT : mouseb & 2 ? B_RIGHT : mouseb & 4 ? B_MIDDLE : 0));
						ev = 1;
					}
				}
				break;
			}
			last_x = mousex;
			last_y = mousey;
			if(!mouseb)
				last_b = mouseb;
		}
		/* Timer event ? */
		if(retval & MU_TIMER)
		{
#ifdef HACK_TEXT
			if(gem_fonts)
			{
				static last_dither_letters = -1;
				if(!ev && (dither_letters != last_dither_letters))
				{
					for(count = 0; count < MAX_DEV; count++)
					{
						struct graphics_device *dev = gem_dev[count];
						if(dev == NULL)
							continue;
						dev->keyboard_handler(dev, 'R', KBD_CTRL); /* reload */
						ev = 1;
					}
					last_dither_letters = dither_letters;
				}
			}
#endif
			if(timeout != NULL)
				timeout_ms -= (DELAY_LOOP_MS/1000); /* evnt_multi delay loop */
		}
		if(timeout != NULL)
			timeout_ms -= (DELAY_LOOP_MS/1000);   /* select delay loop */
		if(ev)
			timeout_ms = 0; /* exit main loop */
	}
	while(((retval = select(n, readfds, writefds, exceptfds, &time_out)) == 0)
	 && (timeout_ms > 0));
	if(ev)
		check_bottom_halves();
	if(retval > 0)
	{
		for(count = 0; count < n; count++)
		{
			if((readfds && FD_ISSET(count, readfds))
			 || (writefds && FD_ISSET(count, writefds)))
				fcntl(count, F_SETFL, 0); /* blocking socket, else sometimes infinite EAGAIN on write and read ??? */
		}
	}
#ifdef DEBUG_GEM
	{
		int rd = 0, wr = 0, ex = 0;
		if(retval >= 0)
			errno = 0;
		for(count = 0; count < n; count++)
		{
			if(readfds && FD_ISSET(count, readfds))
				rd++;
			if(writefds && FD_ISSET(count, writefds))
				wr++;
			if(exceptfds && FD_ISSET(count, exceptfds))
				ex++;
		}
		printf("gem_select retval: %d, errno: %d, rd: %d, wr: %d, ex: %d\r\n", retval, errno, rd, wr, ex);
	}
#endif
	wind_update(BEG_UPDATE);
	return(retval);
}

/* This is a nasty hack used by osdep.c */
#undef get_clipboard_text
#undef set_clipboard_text
#undef clipboard_support

/* clipboard -> links */
unsigned char *gem_get_clipboard_text(struct terminal *term)
{
	short handle;
	char *buf, *p = NULL;
	struct conv_table *ct;
	if(drv != &gem_driver)
		return(stracpy(clipboard));
	if((handle = Fopen(path_scrp, 0)) >= 0)
	{
		long size;
		buf = NULL;
		if(((size = Fseek(0L, handle, 2)) >= 0)
		 && (Fseek(0L, handle, 0) >= 0)
		 && ((buf = (char *)mem_alloc(size+1)) != NULL))
		{
			if((size = Fread(handle, size, buf)) > 0)
			{
				int i;
				buf[size] = '\0';
				while(buf[i])
				{
					buf[i] = (char)atari_to_iso_8859_1(buf[i] & 0xFF);
					i++;
				}
				ct = get_translation_table(get_cp_index("ISO-8859-1"), utf8_table);
				p = (char *)convert_string(ct, (unsigned char *)buf, strlen(buf), NULL);
			}
		}
		if(buf != NULL)
			mem_free(buf);
    Fclose(handle);
	}
	if(p != NULL)
		return(p);
	return(stracpy(clipboard));
}

/* links -> clipboard */
void gem_set_clipboard_text(struct terminal *term, unsigned char *data)
{
	short handle;
	char *buf;
	if(clipboard != NULL)
		mem_free(clipboard);
	if(drv != &gem_driver)
	{
		clipboard = stracpy(data);
		return;
	}
	buf = stracpy(data);
	if(buf != NULL)
	{
		gem_convert_text(buf);
		if((handle = Fcreate(path_scrp, 0)) >= 0)
		{
			Fwrite(handle, strlen(buf), buf);
			Fclose(handle);
		}
		else
			clipboard = stracpy(data);
		mem_free(buf);
	}
}

int gem_clipboard_support(struct terminal *term)
{
	if(drv != &gem_driver)
		return(0);
	return(1);
}

#ifdef HACK_BUTTONS

/* This is a nasty hack used by bfu.c */
#undef display_dlg_item

void gem_display_dlg_item(struct dialog_data *dlg, struct dialog_item_data *di, int sel)
{
	struct terminal *term = dlg->win->term;
	struct graphics_device *dev = term->dev;
	if((F) && (drv == &gem_driver) && (dev != NULL) && (dev->driver_data != NULL))
	{
		struct gem_window *view = (struct gem_window *)dev->driver_data;
		char *text, *text2;
		struct rect rr;
		switch(di->item->type)
		{
			case D_CHECKBOX:
				if(view->iconified)
					return;
				if(!dlg->s)
					restrict_clip_area(dev, &rr, dlg->rr.x1, dlg->rr.y1, dlg->rr.x2, dlg->rr.y2);
				di->l = g_text_width(bfu_style_bw, di->item->gid ? G_DIALOG_RADIO_L G_DIALOG_RADIO_X G_DIALOG_RADIO_R
				 : G_DIALOG_CHECKBOX_L G_DIALOG_CHECKBOX_X G_DIALOG_CHECKBOX_R);
				rs_tedinfo[0].te_ptext = " ";
				if(gem_test_win_top(dev, di->x, di->y, di->x + di->l, di->y + bfu_style_bw->height))
				{
					if(bfu_style_bw->height > gr_hhchar)
						rs_tedinfo[0].te_font = IBM;
					else
						rs_tedinfo[0].te_font = SMALL;
					rs_object[0].ob_x = (short)di->x + view->x + w_border_3d;
					rs_object[0].ob_y = (short)di->y + view->y + h_border_3d;
					rs_object[0].ob_width = (short)di->l - (w_border_3d * 2);
					rs_object[0].ob_height = (short)bfu_style_bw->height - (h_border_3d * 2);
					rs_object[0].ob_spec.tedinfo = rs_tedinfo;
					rs_object[0].ob_state = sel ? (di->checked ? OS_CROSSED | OS_SELECTED : OS_SELECTED) : (di->checked ? OS_CROSSED : OS_NORMAL);
					objc_draw(rs_object, 0, 1, view->clip.x, view->clip.y, view->clip.w, view->clip.h);
				}
				if(sel)
					set_window_ptr(dlg->win, di->x, di->y + G_BFU_FONT_SIZE);
				if(dlg->s)
					exclude_from_set(&dlg->s, di->x, di->y, di->x + di->l, di->y + G_BFU_FONT_SIZE);
				else
					drv->set_clip_area(dev, &rr);
				return;
			case D_BUTTON:
				if(view->iconified)
					return;
				if(!dlg->s)
					restrict_clip_area(dev, &rr, dlg->rr.x1, dlg->rr.y1, dlg->rr.x2, dlg->rr.y2);
				text = _(di->item->text, term);
				text2 = (unsigned char *)mem_alloc(strlen(text) + 5);
				if(text2 == NULL)
					return;
				strcpy(text2, G_DIALOG_BUTTON_L);
				strcat(text2, text);
				strcat(text2, G_DIALOG_BUTTON_R);
				di->l = g_text_width(bfu_style_bw, text2);
				if(gem_test_win_top(dev, di->x, di->y, di->x + di->l, di->y + bfu_style_bw->height))
				{
					strcpy(text2, text);
					gem_convert_text(text2);
					rs_tedinfo[0].te_ptext = text2;
					if(bfu_style_bw->height > gr_hhchar)
						rs_tedinfo[0].te_font = IBM;
					else
						rs_tedinfo[0].te_font = SMALL;
					rs_object[0].ob_x = (short)di->x + view->x + w_border_3d;
					rs_object[0].ob_y = (short)di->y + view->y + h_border_3d;
					rs_object[0].ob_width = (short)di->l - (w_border_3d * 2);
					rs_object[0].ob_height = (short)bfu_style_bw->height - (h_border_3d * 2);
					rs_object[0].ob_spec.tedinfo = rs_tedinfo;
					rs_object[0].ob_state = sel ? OS_SELECTED : OS_NORMAL;
					objc_draw(rs_object, 0, 1, view->clip.x, view->clip.y, view->clip.w, view->clip.h);
				}
				mem_free(text2);
				if(sel)
					set_window_ptr(dlg->win, di->x, di->y + G_BFU_FONT_SIZE);
				if(dlg->s)
					exclude_from_set(&dlg->s, di->x, di->y, di->x + di->l, di->y + G_BFU_FONT_SIZE);
        else
					drv->set_clip_area(dev, &rr);
				return;
			default:
				break;
		}
	}
	display_dlg_item(dlg, di, sel);
}

/* This is a nasty hack used by view_gr.c */
#undef g_text_draw

void gem_text_draw(struct f_data_c *fd, struct g_object_text *t, int x, int y)
{
	struct form_control *form;
	struct form_state *fs;
	struct link *link = t->link_num >= 0 ? fd->f_data->links + t->link_num : NULL;
	struct graphics_device *dev = fd->ses->term->dev;
	if((drv == &gem_driver) && (dev != NULL) && (dev->driver_data != NULL)
	 && link && ((form = link->form)) && ((fs = find_form_state(fd, form))))
	{
		struct gem_window *view = (struct gem_window *)dev->driver_data;
		struct style *inv;
		int in;
		switch(form->type)
		{
			case FC_RADIO:
			case FC_CHECKBOX:
				if(view->iconified)
					return;
				if(link && fd->active && fd->vs->g_display_link && fd->vs->current_link == link - fd->f_data->links)
				{
					inv = g_invert_style(t->style);
					in = 1;
				}
				else
				{
					inv = t->style;
					in = 0;
				}
				if(gem_test_win_top(dev, x, y, x + t->xw, y + inv->height))
				{
					rs_tedinfo[0].te_ptext = "";
					if(inv->height > gr_hhchar)
						rs_tedinfo[0].te_font = IBM;
					else
						rs_tedinfo[0].te_font = SMALL;
					rs_object[0].ob_x = (short)x + view->x + w_border_3d;
					rs_object[0].ob_y = (short)y + view->y + h_border_3d;
					rs_object[0].ob_width = (short)t->xw - (w_border_3d * 2);
					rs_object[0].ob_height = (short)inv->height - (h_border_3d * 2);
					rs_object[0].ob_spec.tedinfo = rs_tedinfo;
					rs_object[0].ob_state = in ? (fs->state ? OS_CROSSED | OS_SELECTED : OS_SELECTED) : (fs->state ? OS_CROSSED : OS_NORMAL);
					objc_draw(rs_object, 0, 1, view->clip.x, view->clip.y, view->clip.w, view->clip.h);
				}
				if(in)
					g_free_style(inv);
				return;
			case FC_SELECT:
//	=> remove 2 put_chrs("[",... put_chrs("]",... inside html.c for a better look
				if(view->iconified)
					return;
				if(link && fd->active && fd->vs->g_display_link && fd->vs->current_link == link - fd->f_data->links)
				{
					inv = g_invert_style(t->style);
					in = 1;
				}
				else
				{
					inv = t->style;
					in = 0;
				}
				fixup_select_state(form, fs);
				if(gem_test_win_top(dev, x, y, x + t->xw, y + inv->height))
				{
					unsigned char *text = (unsigned char *)mem_alloc(strlen(t->text) + 1);
					if(text != NULL)
					{
						strcpy(text, fs->state < form->nvalues ? (char *)form->labels[fs->state] : " ");
						gem_convert_text(text);
						rs_tedinfo[0].te_ptext = text;
						if(inv->height > gr_hhchar)
							rs_tedinfo[0].te_font = IBM;
						else
							rs_tedinfo[0].te_font = SMALL;
						rs_object[0].ob_x = (short)x + view->x + w_border_3d;
						rs_object[0].ob_y = (short)y + view->y + h_border_3d;
						rs_object[0].ob_width = (short)t->xw - (w_border_3d * 2);
						rs_object[0].ob_height = (short)inv->height - (h_border_3d * 2);
						rs_object[0].ob_spec.tedinfo = rs_tedinfo;
						rs_object[0].ob_state = in ? OS_SELECTED : OS_NORMAL;
						objc_draw(rs_object, 0, 1, view->clip.x, view->clip.y, view->clip.w, view->clip.h);
						mem_free(text);
					}
				}
				if(in)
					g_free_style(inv);
				return;
			case FC_SUBMIT:
			case FC_RESET:
			case FC_BUTTON:
				if(view->iconified)
					return;
				if(link && fd->active && fd->vs->g_display_link && fd->vs->current_link == link - fd->f_data->links)
				{
					inv = g_invert_style(t->style);
					in = 1;
				}
				else
				{
					inv = t->style;
					in = 0;
				}
				if(gem_test_win_top(dev, x, y, x + t->xw, y + inv->height))
				{
					unsigned char *text = (unsigned char *)mem_alloc(strlen(t->text) + 1);
					if(text != NULL)
					{
						if((t->text[0] == '[') && (t->text[strlen(t->text) - 1] == ']'))
						{
							strcpy(text, &t->text[1]);
							text[strlen(t->text) - 2] = 0;
						}							
						else
							strcpy(text, t->text);
						gem_convert_text(text);
						rs_tedinfo[0].te_ptext = text;
						if(inv->height > gr_hhchar)
							rs_tedinfo[0].te_font = IBM;
						else
							rs_tedinfo[0].te_font = SMALL;
						rs_object[0].ob_x = (short)x + view->x + w_border_3d;
						rs_object[0].ob_y = (short)y + view->y + h_border_3d;
						rs_object[0].ob_width = (short)t->xw - (w_border_3d * 2);
						rs_object[0].ob_height = (short)inv->height - (h_border_3d * 2);
						rs_object[0].ob_spec.tedinfo = rs_tedinfo;
						rs_object[0].ob_state = in ? OS_SELECTED : OS_NORMAL;
						objc_draw(rs_object, 0, 1, view->clip.x, view->clip.y, view->clip.w, view->clip.h);
						mem_free(text);
					}
				}
				if(in)
					g_free_style(inv);
				return;	
			default:
				break;
		}
	}
	g_text_draw(fd, t, x, y);
}

#else /* !HACK_BUTTONS */

/* This is a nasty hack used by buf.c */
#undef display_dlg_item

void gem_display_dlg_item(struct dialog_data *dlg, struct dialog_item_data *di, int sel)
{
	display_dlg_item(dlg, di, sel);
}

/* This is a nasty hack used by view_gr.c */
#undef g_text_draw

void gem_text_draw(struct f_data_c *fd, struct g_object_text *t, int x, int y)
{
	g_text_draw(fd, t, x, y);
}

#endif /* HACK_BUTTONS */

/* This is a nasty hack used by dip.c */
#undef g_text_width
#undef g_char_width
#undef g_print_text
#undef g_wrap_text

#ifdef HACK_TEXT

static unsigned char *make_font_name(int attr)
{
	int len = 0;
	unsigned char *str = init_str();
	add_to_str(&str, &len, G_HTML_DEFAULT_FAMILY);
	add_to_str(&str, &len, attr & AT_BOLD ? "-bold" : "-medium");
	add_to_str(&str, &len, attr & AT_ITALIC ? "-italic-serif" : "-roman-serif");
	add_to_str(&str, &len, attr & AT_FIXED ? "-mono" : "-vari");
	return(str);
}

int gem_text_width(struct style *style, unsigned char *text)
{
	static struct style *last_style = NULL;
	static struct style last;
	if(!dither_letters && gem_fonts && (drv == &gem_driver))
	{
		short ext[8], dummy;
		unsigned char *p, *p2;
		unsigned char *buf = (unsigned char *)stracpy(text);
		if(buf != NULL)
		{
			short id = 1, vector = 0;
			long h_factor = 0, w_factor = 0;
			int w;
			int len = gem_convert_text(buf);
			if(!len)
			{
				mem_free(buf);
				return(0);
			}
			if((last_style == NULL) || (style != last_style)
			 || (last.height != style->height) || (last.flags != style->flags)
			 ||	(last.r0 != style->r0) || (last.r1 != style->r1))
			{	
				struct gem_fontinfo *pfont = gem_pfont;
				while(pfont != NULL)
				{
					if((strcmp(style->font, pfont->name) == 0)
					 && !(!pfont->vector && (style->height > 18)))
					{
						if(id <= 1)
						{
							id = pfont->id;
							vector = pfont->vector;
							h_factor = pfont->h_factor;
							w_factor = pfont->w_factor;
						}
						else if(pfont->default_font)
						{
							id = pfont->id;
							vector = pfont->vector;
							h_factor = pfont->h_factor;
							w_factor = pfont->w_factor;
							break;
						}
					}
					pfont = pfont->next;
				}
				if((id <= 1) && ((p = strchr(style->font, '-')) != NULL))
				{ /* not found => second pass without font name */
					pfont = gem_pfont;
					while(pfont != NULL)
					{
						if(((p2 = strchr(pfont->name, '-')) != NULL)
						 && (strcmp(p, p2) == 0)
						 && !(!pfont->vector && (style->height > 18)))
						{
							if(id <= 1)
							{
								id = pfont->id;
								vector = pfont->vector;
								h_factor = pfont->h_factor;
								w_factor = pfont->w_factor;
							}
							else if(pfont->default_font)
							{
								id = pfont->id;
								vector = pfont->vector;
								h_factor = pfont->h_factor;
								w_factor = pfont->w_factor;
								break;
							}
						}		
						pfont = pfont->next;
					}
				}
				vst_font(vdi_handle, id);
				vst_charmap(MAP_ATARI);
				vst_rotation(vdi_handle, 0);
				vst_effects(vdi_handle, style->flags ? TXT_UNDERLINED : TXT_NORMAL);
				vst_alignment(vdi_handle, TA_LEFT, TA_ASCENT, &dummy, &dummy);
				if(vector)
				{
					vst_arbpt(vdi_handle, (short)(((long)style->height * h_factor) / 1000L), &dummy, &dummy, &dummy, &dummy);
					vst_setsize(vdi_handle, (short)(((long)g_char_width(style, 'W') * w_factor) / 1000L), &dummy, &dummy, &dummy, &dummy);
				}
				else
				{
					short min_char, max_char, distances[5], maxwidth, effects[3];
					vqt_fontinfo(vdi_handle, &min_char, &max_char, distances, &maxwidth, effects);
					h_factor = ((long)distances[4] * 1000L) / (long)(distances[0] + distances[4]);
					vst_height(vdi_handle, (short)(((long)style->height * h_factor) / 1000L), &dummy, &dummy, &dummy, &dummy);
				}
			}
#if 1
			vqt_extent(vdi_handle, (char *)buf, ext);
			w = (int)(ext[2] - ext[0]);
#else
			w = 0;
			p = buf;
			while(*p)
			{
				short width;
				if(vqt_width(vdi_handle, (short)*p & 0xFF, &width, &dummy, &dummy) != -1)
					w += width;
				p++;
			}
#endif
			mem_free(buf);
			last_style = style;
			last.r0 = style->r0;
			last.r1 = style->r1;
			last.height = style->height;
			last.flags = style->flags;
			return(w);
		}
  }
  last_style = NULL;
	return(g_text_width(style, text));
}

int gem_char_width(struct style *style, int charcode)
{
	if(!dither_letters && gem_fonts && (drv == &gem_driver))
		return(gem_text_width(style, encode_utf_8(charcode)));
	return(g_char_width(style, charcode));
}

void gem_print_text(struct graphics_driver *gd, struct graphics_device *dev, int x, int y, struct style *style, unsigned char *text, int *width)
{
	if(!dither_letters && gem_fonts && (gd == &gem_driver) && (dev != NULL) && (dev->driver_data != NULL))
	{
		struct gem_window *view = (struct gem_window *)dev->driver_data;
		short rect[4], dummy;
		int color, w;
		unsigned char *buf;
		if(view->iconified)
			return;
		w = gem_text_width(style, text);
		if(width != NULL)
			*width += w;
		if((y + style->height <= dev->clip.y1) || (y >= dev->clip.y2))
		 	return;
		if(!gem_test_win_top(dev, x, y, x + w, y + style->height))
			return;
		buf = (unsigned char *)stracpy(text);
		if(buf != NULL)
		{
			gem_convert_text(buf);
			color = (int)(((unsigned long)style->r0 << 16) + ((unsigned long)style->g0 << 8) + (unsigned long)style->b0);
			vsf_color(vdi_handle, (short)dip_get_color_sRGB(color));
			vsf_interior(vdi_handle, 1);
			vsf_perimeter(vdi_handle, 0);
			color = (int)(((unsigned long)style->r1 << 16) + ((unsigned long)style->g1 << 8) + (unsigned long)style->b1); 
			vst_color(vdi_handle, (short)dip_get_color_sRGB(color));
			vswr_mode(vdi_handle, MD_REPLACE);
			v_hide_c(vdi_handle);
			rect[0] = (short)x + view->x;
			rect[1] = (short)y + view->y;
			rect[2] = rect[0] + w - 1;
			rect[3] = rect[1] + style->height - 1;
			v_bar(vdi_handle, rect);
			vswr_mode(vdi_handle, MD_TRANS);
			v_gtext(vdi_handle, (short)x + view->x, (short)y + view->y, (char *)buf);
			v_show_c(vdi_handle, 1);
			mem_free(buf);
			return;
		}
	}
	g_print_text(gd, dev, x, y, style, text, width);
}

int gem_wrap_text(struct wrap_struct *w)
{
	if(!dither_letters && gem_fonts && (drv == &gem_driver))
	{
		int sm = gem_char_width(w->style, '-');
		while(*w->text)
		{
			int u, s;
			if(*w->text == ' ')
				w->last_wrap = w->text,
			w->last_wrap_obj = w->obj;
			GET_UTF_8(w->text, u);
			if(!u)
				continue;
			if(u == 0x01 || u == 0xa0)
				u = ' ';
			if(u == 0xad)
				s = 0;
			else
				s = gem_char_width(w->style, u);
			if((w->pos += s) <= w->width)
			{
				c:
				if(u != 0xad || *w->text == ' ')
					continue;
				if(w->pos + sm <= w->width || (!w->last_wrap && !w->last_wrap_obj))
				{
					w->last_wrap = w->text;
					w->last_wrap_obj = w->obj;
					continue;
				}
			}
			if(!w->last_wrap && !w->last_wrap_obj)
				goto c;
			return(0);
		}
		return(1);
  }
	return(g_wrap_text(w));
}

#else /* !HACK_TEXT */

int gem_text_width(struct style *style, unsigned char *text)
{
	return(g_text_width(style, text));
}

int gem_char_width(struct style *style, int charcode)
{
	return(g_char_width(style, charcode));
}

void gem_print_text(struct graphics_driver *gd, struct graphics_device *dev, int x, int y, struct style *style, unsigned char *text, int *width)
{
	g_print_text(gd, dev, x, y, style, text, width);
}

int gem_wrap_text(struct wrap_struct *w)
{
	return(g_wrap_text(w));
}

#endif /* HACK_TEXT */

static unsigned char *gem_init_driver(unsigned char *param, unsigned char *display)
{
	COOKIE *p;
	char message[128];
	short work_in[12], work_out[272], dummy;
	int i, j;
	long colors;
	for(i = 0; i < g_argc; i++)
	{
		char *p = g_argv[i];
		if((strlen(p) > 3)
		 && (p[0] >= 'A') && (p[0] <= 'Z')
		 && (p[1] == ':') && (p[2] == '\\'))
		{
			j = 2;
			while(p[j])
			{
				if(p[j] == '\\')
					p[j] = '/';
				j++;
			}
		}		
	}
#ifdef DEBUG_GEM
	printf("gem_init_driver appl_init()\r\n");
#endif
	memset(gem_dev, 0, sizeof(struct graphics_device *) * MAX_DEV);
	/* Open AES (Application Environment Services) */
	if((ap_id = appl_init()) == -1)
	{
		snprintf(message, sizeof (message), "Can not open AES\n");
		return(stracpy(message));
	}
	if((_AESversion < 0x330)
	 || !objc_sysvar(0, AD3DVAL, 2, 2, &w_border_3d, &h_border_3d))
		w_border_3d	= h_border_3d = 2;
#ifdef DEBUG_GEM
	printf("gem_init_driver graf_handle()\r\n");
#endif
	vdi_handle = graf_handle(&gr_hwchar, &gr_hhchar, &dummy, &dummy);
	if(vdi_handle < 1)
	{
		snprintf(message, sizeof (message), "Wrong VDI handle %d returned by AES\n", vdi_handle);
		return(stracpy(message));
	}
#ifdef DEBUG_GEM
	printf("gem_init_driver v_opnvwk()\r\n");
#endif
	/* Open virtual VDI workstation */
	work_in[0] = Getrez() + 2;
	for(i = 1; i < 10; i++)
		work_in[i] = 1;
	work_in[10] = 2;
	v_opnvwk(work_in, &vdi_handle, work_out);
	if(!vdi_handle)
	{
		snprintf(message, sizeof (message), "Can not open VDI virtual workstation\n");
		return(stracpy(message));
	}
#ifdef HACK_TEXT
	gem_fonts = work_out[10];
#endif
#ifdef DEBUG_GEM
	printf("gem_init_driver wind_get(DESKTOP_HANDLE,WF_WORKXYWH)\r\n");
#endif
	/* Read desktop size and position */
	width_screen = work_out[0] + 1;
	height_screen = work_out[1] + 1;
	if(!wind_get(DESKTOP_HANDLE, WF_WORKXYWH, &desk_x, &desk_y, &desk_w, &desk_h))
	{
		snprintf(message, sizeof (message), "Can not read desktop properties\n");
		return(stracpy(message));
	}
#ifdef DEBUG_GEM
	printf("gem_init_driver vq_extnd()\r\n");
#endif
	/* Read bit depth */
	vq_extnd(vdi_handle, 1, work_out);
	bpp = work_out[4];
	packed_format = 0; /* interleaved */
	little_endian = 0; /* big endian */
	colors = (1 << bpp);
	p = get_cookie(0x45644449L);  /* EdDI */
	if(p != NULL)
	{
#ifdef DEBUG_GEM
		printf("gem_init_driver vq_scrninfo()\r\n");
#endif
		vq_scrninfo(vdi_handle, work_out);
		if(work_out[0] == 2) /* format */
			packed_format = 1; /* packed */
		if(Atari_get_EdDI_version((void *)p->v.l) >= 0x110)
		{
			if(work_out[14] & 0x80) /* bit organization */
				little_endian = 1;
			if((bpp == 16) && !(work_out[14] & 0x02))
				colors = *((long *)&work_out[3]);
			else if(bpp != 16)
				colors = *((long *)&work_out[3]);
		}
	}
	/* depth: <-B8:bitorder, B7-B3:bpp, B2-B0:pixelsize */
	switch(bpp)
	{
		case 4:          /* 0x21 */
			gem_driver.depth = 33;	    /* 4 bpp, 1Bpp, color 121 */
			pixelsize = 1;
			break;
		case 8:          /* 0x41 */
			gem_driver.depth = 65;	    /* 8 bpp, 1Bpp, color 332 */
			pixelsize = 1;
			break;
		case 15:
		case 16:
			if(colors == 32768) /* 0x7A */
				gem_driver.depth = 122;   /* 15 bpp, 2Bpp, color 555  */
			else
			{
				if(little_endian) /* 0x82 */
					gem_driver.depth = 130; /* 16 bpp, 2Bpp, color 565 */
				else             /* 0x182 */
					gem_driver.depth = 386; /* 16 bpp, 2Bpp, color 565 BE */
			}
			pixelsize = 2;
			break;
		case 24:
			if(little_endian) /* 0xC3 */
				gem_driver.depth = 195;   /* 24 bpp, 3Bpp, color BGR */
			else             /* 0x1C3 */
				gem_driver.depth = 451;   /* 24 bpp, 3Bpp, color RGB */
			pixelsize = 3;
			break;
		case 32:
			if(little_endian) /* 0xC4 */
				gem_driver.depth = 196;   /* 24 bpp, 4Bpp, color BGR0 */
			else             /* 0x2C4 */
				gem_driver.depth = 708;   /* 24 bpp, 4Bpp, color 0RGB */
			pixelsize = 4;
			break;
		default:
			form_alert(1, "[1][Colour depth not supported!][Cancel]");
			snprintf(message, sizeof(message), "%d bits colour depth not supported\n", bpp);		
			return(stracpy(message));
	}
	/* buttons bar */
	if(!(rsrc_load("ilinks.rsc") && rsrc_gaddr(R_TREE, FORM1, &buttons)))
		buttons = NULL;
	/* clipboard init */
	j = 1;
	if(scrp_read(path_scrp) && path_scrp[0])
	{
		char *end_path = path_scrp;
		j = 0;
		for(i=0; path_scrp[i] && i < MAX_PATH; i++)
		{
			if(path_scrp[i]=='\\')
				end_path = &path_scrp[i];
		}
		*end_path = '\0';
	}
	else
	{
		void *save_ssp = (void *)Super(0L);
		path_scrp[0] = (char)*_bootdev + 'A';
		Super(save_ssp);
		strcat(&path_scrp[1], ":\\clipbrd");
	}
	if(Fsfirst(path_scrp, 0x10) < 0)
		Dcreate(path_scrp);
	strcat(path_scrp, "\\");
	if(j)
		scrp_write(path_scrp);
	strcat(path_scrp, "scrap.txt");
#ifdef HACK_TEXT
	/* GEM fonts init */
	gem_pfont = NULL;
	if(vq_vgdos() != -2)
	{
		short index, id, i, w, vector, cell_height, min_char, max_char, monospace, w_mini, w_maxi;
		short distances[5], maxwidth, effects[3];
		static char name[34];
		struct gem_fontinfo *pfont;
		gem_fonts += vst_load_fonts(vdi_handle, 0);
#ifdef DEBUG_GEM
		printf("gem_init_driver gem_fonts: %d\r\n", gem_fonts);
#endif
		vst_scratch(vdi_handle, SCRATCH_BOTH);
		vst_kern(vdi_handle, TRACK_NONE, PAIR_OFF, &dummy, &dummy);	
		for(index = 1; index <= gem_fonts; index++)
		{
#define FONT_SIZE 100L
			struct gem_fontinfo *p;
			id = vqt_name(vdi_handle, index, name);
			vector = (short)name[32] & 0xFF;
			name[32] = '\0';
			vst_font(vdi_handle, id);
			if(vector)
				vst_arbpt(vdi_handle, FONT_SIZE, &dummy, &dummy, &dummy, &cell_height);
			else
				vst_point(vdi_handle, FONT_SIZE, &dummy, &dummy, &dummy, &cell_height);
			vqt_fontinfo(vdi_handle, &min_char, &max_char, distances, &maxwidth, effects);
			w_mini = 32767;
			w_maxi = 0;
			for(i = min_char; i <= max_char; i++)
			{
				if(vqt_width(vdi_handle, i, &w, &dummy, &dummy) != -1)
				{
					if(w < w_mini)
						w_mini = w;
					if(w > w_maxi)
						w_maxi = w;
				}				
			}
			if(w_maxi == w_mini)
				monospace = 1;
			else
				monospace = 0;
#ifdef DEBUG_GEM
			printf("gem_init_driver font_ID: %d, name: %s, vector: %d, monospace: %d\r\n", id, name, vector, monospace);
#endif
			p = (struct gem_fontinfo *)mem_alloc(sizeof(struct gem_fontinfo));
			if(p != NULL)
			{
				int default_font = 0;
				int attr = monospace ? AT_FIXED : 0;
				if(strstr(name, "Bold") != NULL)
					attr |= AT_BOLD;				
				if(strstr(name, "Italic") != NULL)
					attr |= AT_ITALIC;
				if((strstr(name, "Courrier") != NULL)
				 || (strstr(name, "Swiss") != NULL))     /* Speedo 4 */
				 	default_font = 1;
				if(((strstr(name, "Monospace ") != NULL) /* Speedo 4 */
				 || (strstr(name, "Monaco") != NULL)) && (attr == AT_FIXED))
					p->name = stracpy("monospaced");
				else if((strstr(name, "Monaco") != NULL) && (attr == 0)) 				
					p->name = stracpy(G_BFU_DEFAULT_FONT);
				else if(!vector)
				{
					mem_free(p);
					continue;
				}
				else
					p->name = make_font_name(attr);
				if(p->name == NULL)
					mem_free(p);
				else
				{
					p->next = NULL;
					p->id = id;
					p->vector = vector;
					p->monospace = monospace;
					p->h_factor = (FONT_SIZE * 1000L) / (long)cell_height;
					vqt_width(vdi_handle, 'W', &w_maxi, &dummy, &dummy);
					p->w_factor = (FONT_SIZE * 1000L) / (long)w_maxi;
					p->attr = attr;
					p->default_font = default_font;
					if(gem_pfont == NULL)
						pfont = gem_pfont = p;
					else
					{
						pfont->next = p;
						pfont = p;
					}
				}
			}
		}
		if(gem_pfont == NULL)
		{
			vst_unload_fonts(vdi_handle, 0);
			gem_fonts = 0;
		}	
	}
	else
		gem_fonts = 0;
#endif
	wind_update(BEG_UPDATE);
	/* palette init */
	if(pixelsize == 1)
	{
		short *rgb = save_palette;
		int max_col = (bpp < 8) ? 16 : 256;
		int c;
		for(c = 0; c < max_col; c++, rgb += 3)
			vq_color(vdi_handle, c, 0, rgb);
	}
	memset(tab_rgb,  0 , 256 * sizeof(long));
	if(pixelsize == 1)
	{
		short rgb[3]; 
		int c, cr, r, g, b;
		if(bpp == 4) /* 16 colors */
		{
			unsigned short *p = Atari_table_c2p;
			for(i = 0; i < 16; i++)
			{
				for(c = 0; c < 16; c++)
				{		
					short index = tab_index_png_vdi[c];
					unsigned char color = tab_index_col_pal[index];
					*p++ = (color & 1) ? 1 : 0;
					*p++ = (color & 2) ? 1 : 0;
					*p++ = (color & 4) ? 1 : 0;
					*p++ = (color & 8) ? 1 : 0;
					r = (c & 8 ) ? 255 : 0;
					g = ((c >> 1) & 3) * 85;
					b = (c & 1) ? 255 : 0;
					tab_rgb[index] = (r << 16) + (g << 8) + b;
					rgb[0] = (short)((r * 1001) >> 8);
					rgb[1] = (short)((g * 1001) >> 8);
					rgb[2] = (short)((b * 1001) >> 8);
					vs_color(vdi_handle, index, rgb);		
				}
			}
		}
		else /* 256 colors */
		{
			/* 256 to 64 bmp color index */
			for(c = 0; c < 256; c++)
			{
				cr = (((c >> 6) & 3) << 4) + (((c >> 3) & 3) << 2) + (c & 3) + 16;
				tab_index_col_png[c] = (unsigned char)cr;
				if(!tab_rgb[cr])
				{
					r = reduce_palette[(c >> 5) & 7] * 36;
					g = reduce_palette[(c >> 2) & 7] * 36;
					b = (c & 3) * 84;
					tab_rgb[cr] = (r << 16) + (g << 8) + b;
					rgb[0] = (short)((r * 1001) >> 8);
					rgb[1] = (short)((g * 1001) >> 8);
					rgb[2] = (short)((b * 1001) >> 8);
					vs_color(vdi_handle, (short)cr, rgb);
				}
			}
		}
	}
	icon = (unsigned char *)mem_alloc(links_image.width * links_image.height * pixelsize);
	if(icon != NULL)
	{
		unsigned char *src = links_image.pixel_data;
		unsigned char *dst = icon;
		unsigned short *dstw = (unsigned short *)icon;
		for(j = 0; j < links_image.height; j++)
		{
			switch(bpp)
			{
				case 4:
					for(i = 0; i < links_image.width; i++)
					{
						unsigned char color;
						color = src[0] ? 8 : 0;      /* red */
						color |= (src[1] >> 6) << 1; /* green */
						color |= (src[2] ? 1 : 0);   /* blue */
						*dst++ = color;
						src += links_image.bytes_per_pixel;
					}
					break;
				case 8:
					for(i = 0; i < links_image.width; i++)
					{
						unsigned long color;
						color = (unsigned long)src[0];        /* red */
						color <<= 8;
						color |= (unsigned long)src[1];       /* green */
						color <<= 8;
						color |= (unsigned long)src[2];       /* blue */
						switch(color)
						{
							case 0x000000: *dst++ = 255; break; /* black palette index */
							case 0xFF0000: *dst++ = 1; break;   /* red palette index */
							case 0x00FF00: *dst++ = 2; break;   /* green palette index */
							case 0x0000FF: *dst++ = 4; break;   /* blue palette index */
							default: *dst++ = 0; break;         /* white palette index */
						}
						src += links_image.bytes_per_pixel;
					}
					break;
				case 15:
				case 16:
					for(i = 0; i < links_image.width; i++)
					{
						unsigned short color;
						if(colors == 32768)
						{
							color = (unsigned short)src[0] << 7;
							color &=  0x7C00;
							color |= ((unsigned short)src[1] << 2);
							color &=  0x7FE0;
							color |= ((unsigned short)src[2] >> 3);
						}
						else
						{
							color = (unsigned short)src[0] << 8;
							color &=  0xF800;
							color |= ((unsigned short)src[1] << 3);
							color &=  0xFFE0;
							color |= ((unsigned short)src[2] >> 3);
							if(little_endian)
								color = (color >> 8) | (color << 8);
						}
						*dstw++ = color;
						src += links_image.bytes_per_pixel;
					}
					break;
				case 24:
					for(i = 0; i < links_image.width; i++)
					{
						if(little_endian)
						{
							*dst++ = src[2];
							*dst++ = src[1];
							*dst++ = src[0];
						}
						else
						{
							*dst++ = src[0];
							*dst++ = src[1];
							*dst++ = src[2];
						}
						src += links_image.bytes_per_pixel;			
					}
					break;
				case 32:
					for(i = 0; i < links_image.width; i++)
					{
						if(little_endian)
						{
							*dst++ = src[2];
							*dst++ = src[1];
							*dst++ = src[0];
							*dst++ = 0;
						}
						else
						{
							*dst++ = 0;
							*dst++ = src[0];
							*dst++ = src[1];
							*dst++ = src[2];
						}
						src += links_image.bytes_per_pixel;			
					}
					break;
			}
		}
	}	
	return(NULL);
}

static struct graphics_device *gem_init_device()
{
	short rect[4];
	short x, y, w, h;
	short posx, posy, width, height, handle, type;
	struct gem_window *view;
	struct graphics_device *dev;
	int c;
	if(!vdi_handle)
		return(NULL);
	width = (desk_w * 2) /3;
	height = (desk_h * 2) / 3; 
	/* Center our window */
	posx = desk_x;
	posy = desk_y;
	if(width < desk_w)
		posx += (desk_w - width) >> 1;
	if(height < desk_h)
		posy += (desk_h - height) >> 1;
	/* Calculate our window size and position */
	type = NAME | CLOSER | FULLER | MOVER | SIZER | SMALLER;
	if(!wind_calc(0, type, posx, posy, width, height, &x, &y, &w, &h))
		return(NULL);
	if(x < desk_x)
		x = desk_x;
	if(y < desk_y)
		y = desk_y;
#ifdef DEBUG_GEM
	printf("gem_init_device wind_create()\r\n");
#endif
	/* Create window */
	handle = wind_create(type, x, y, w, h);
	if(handle < 0)
	{
#ifdef DEBUG_GEM
		printf("gem_init_device error wind_create!\r\n");
#endif
		return(NULL);
	}
	dev = (struct graphics_device *)mem_alloc(sizeof(struct graphics_device));
	if(dev == NULL)
	{
#ifdef DEBUG_GEM
		printf("gem_init_device error malloc!\r\n");
#endif
		wind_delete(handle);
		return(NULL);
	}
	for(c = 0; c < MAX_DEV; c++)
	{
		if(gem_dev[c] == NULL)
		{
			gem_dev[c] = dev;
			break;
		}
	}
	if(c >= MAX_DEV)
	{
#ifdef DEBUG_GEM
		printf("gem_init_device error MAX_DEV!\r\n");
#endif
		mem_free(dev);
		wind_delete(handle);
		return(NULL);
	}
	view = (struct gem_window *)mem_alloc(sizeof(struct window));
	if(view == NULL)
	{
#ifdef DEBUG_GEM
		printf("gem_init_device error malloc!\r\n");
#endif
		gem_dev[c] = NULL;
		mem_free(dev);
		wind_delete(handle);
		return(NULL);
	}
	view->handle = handle;
	view->type = type;
	view->buttons = 1;
	view->fulled = 0;
	view->flag_redraw = 0;
	view->iconified = 0;
#ifdef DEBUG_GEM
	printf("gem_init_device wind_open()\r\n");
#endif
	/* Open the window */
	wind_open(handle, x, y, w, h);
	wind_get(handle, WF_WORKXYWH, &x, &y, &w, &h);
	dev->drv = &gem_driver;
	dev->driver_data = view;
	/* Clear the window */
	v_hide_c(vdi_handle);
	vsf_color(vdi_handle, 0);
	vsf_interior(vdi_handle, 1);
	vsf_perimeter(vdi_handle, 0);
	vswr_mode(vdi_handle, MD_REPLACE);
	rect[0] = (short)x;
	rect[1] = (short)y;
	rect[2] = (short)(x + w - 1);
	rect[3] = (short)(y + h - 1);
	v_bar(vdi_handle, rect);
	v_show_c(vdi_handle, 1);
	rect[0] = 0;
	rect[1] = 0;
	rect[2] = 1;
	rect[3] = 1;
	vs_clip(vdi_handle, 1, rect);
	gem_get_size(dev);
	return(dev);
}

static void gem_shutdown_device(struct graphics_device *dev)
{
	int c;
	if(dev != NULL)
	{
		struct gem_window *view = (struct gem_window *)dev->driver_data;
		if(view != NULL)
		{
			if(view->handle >= 0)
			{
#ifdef DEBUG_GEM
				printf("gem_shutdown_device wind_close/delete()\r\n");
#endif
				wind_close(view->handle);
				wind_delete(view->handle);
			}
			mem_free(view);
		}
		mem_free(dev);
	}
	for(c = 0; c < MAX_DEV; c++)
	{
		if(gem_dev[c] == dev)
		{
			gem_dev[c] = NULL;
			break;
		}
	}
}

static void gem_shutdown_driver(void)
{
	if(icon != NULL)
		mem_free(icon);
	wind_update(END_UPDATE);
#ifdef DEBUG_GEM
	printf("gem_shutdown_driver v_clsvwk()\r\n");
#endif
	if(vdi_handle)
	{
		short *rgb = save_palette;
		int c;
		if(pixelsize == 1)
		{
			int max_col = (bpp < 8) ? 16 : 256;
			for(c = 0; c < max_col; c++, rgb += 3)
				vs_color(vdi_handle, c, rgb);
		}
#ifdef HACK_TEXT
		if(gem_fonts)
		{
			struct gem_fontinfo *pfont = gem_pfont;
			while(pfont != NULL)
			{
				struct gem_fontinfo *p = pfont;
				if(pfont->name != NULL)
					mem_free(pfont->name);
				pfont = pfont->next;
				mem_free(p);
			}
			vst_unload_fonts(vdi_handle, 0);
		}
#endif
		v_clsvwk(vdi_handle);
	}
#ifdef DEBUG_GEM
	printf("gem_shutdown_driver appl_exit()\r\n");
#endif
	appl_exit();
}

static unsigned char *gem_get_driver_param(void)
{
#ifdef DEBUG_GEM
	printf("gem_get_driver_param\r\n");
#endif
	return(NULL);
}

/* Return value:	0 alloced on heap
 *			1 alloced in vidram
 *			2 alloced in X server shm
 */
static int gem_get_empty_bitmap(struct bitmap *dest)
{
#ifdef DEBUG_GEM
	printf("gem_get_empty_bitmap dest: 0x%X\r\n", dest);
#endif
	dest->data = mem_alloc(((dest->x + 15) & ~15) * dest->y * pixelsize);
	dest->skip = ((dest->x + 15) & ~15) * pixelsize;
	dest->flags = 0;
	return(0);
}

/*
static int gem_get_filled_bitmap(struct bitmap *bmp, long color)
{
#ifdef DEBUG_GEM
	printf("gem_get_filled_bitmap\r\n");
#endif
	return 0;
}
*/

static void gem_register_bitmap(struct bitmap *bmp)
{
#ifdef DEBUG_GEM
	printf("gem_register_bitmap\r\n");
#endif
}

static void *gem_prepare_strip(struct bitmap *bmp, int top, int lines)
{
#ifdef DEBUG_GEM
	printf("gem_prepare_strip\r\n");
#endif
	return(((char *)bmp->data) + bmp->skip * top);
}

static void gem_commit_strip(struct bitmap *bmp, int top, int lines)
{
#ifdef DEBUG_GEM
	printf("gem_commit_strip\r\n");
#endif
}

static void gem_unregister_bitmap(struct bitmap *bmp)
{
#ifdef DEBUG_GEM
	printf("gem_unregister_bitmap\r\n");
#endif
	mem_free(bmp->data);
}

static void gem_draw_bitmap(struct graphics_device *dev, struct bitmap *bmp, int x, int y)
{
	struct gem_window *view;
	MFDB mfdb_src, mfdb_dst;
	short pxy[8];
	int bmp_size;
	void *src = NULL, *dst = NULL;
	if((dev == NULL) || (dev->driver_data == NULL) || !vdi_handle) 
		return;
	if(!bmp->x || !bmp->y)
		return;
	view = (struct gem_window *)dev->driver_data;
	if(view->iconified)
		return;
	if(!gem_test_win_top(dev, x, y, x + bmp->x, y + bmp->y))
		return;
#ifdef DEBUG_GEM
	printf("gem_draw_bitmap x: %d, y: %d\r\n", x, y);
#endif
	pxy[0] = pxy[1] = 0;
	pxy[2] = (short)bmp->x - 1;
	pxy[3] = (short)bmp->y - 1;
	pxy[4] = (short)x + view->x;
	pxy[5] = (short)y + view->y;
	pxy[6] = (short)x + (short)bmp->x + view->x - 1;
	pxy[7] = (short)y + (short)bmp->y + view->y - 1;
	mfdb_src.fd_addr = bmp->data;
	mfdb_src.fd_w = (short)((bmp->x + 15) & ~15); /* alignment */
	mfdb_src.fd_h = (short)bmp->y;
	mfdb_src.fd_nplanes = (short)bpp;
	mfdb_src.fd_wdwidth = mfdb_src.fd_w >> 4;
	mfdb_src.fd_stand = mfdb_src.fd_r1 = mfdb_src.fd_r2 = mfdb_src.fd_r3 = 0;
	memset(&mfdb_dst, 0, sizeof(MFDB));
	bmp_size = ((bmp->x + 15) & ~15) * bmp->y;
	if(pixelsize == 1)
	{
		short rect[4];
		rect[0] = pxy[4];  /* x, y, w, h */
		rect[1] = pxy[5];
		rect[2] = pxy[6] - (pxy[4] - 1);
		rect[3] = pxy[7] - (pxy[5] - 1);
		if(rc_intersect((GRECT *)&view->clip, (GRECT *)rect))
		{
			if((bpp == 4) && !packed_format) /* 16 colors */
			{
				if((dst = mem_alloc(bmp_size)) != NULL)
				{		
					Atari_C2pConvert4(bmp->data, dst, (unsigned long)((bmp->x + 15) & ~15), (unsigned long)bmp->y, 0, (unsigned long)bmp->skip, (unsigned long)((bmp->x + 15) & ~15) >> 1);
					mfdb_src.fd_addr = dst;
				}
			}
			else /* 256 colors */
			{
				if(((src = mem_alloc(bmp_size)) != NULL)
				 && ((dst = mem_alloc(bmp_size)) != NULL))
				{		
					unsigned char *p1 = bmp->data;
					unsigned char *p2 = src;
					bmp_size = bmp->skip * bmp->y; 
					while(bmp_size--)
						*p2++ = tab_index_col_png[*p1++]; /* 256 to 64 colors */
					if(!packed_format)
					{
						Atari_C2pConvert8(src, dst, (unsigned long)((bmp->x + 7) & ~7), (unsigned long)bmp->y, 0, (unsigned long)bmp->skip, (unsigned long)bmp->skip);
						mfdb_src.fd_addr = dst;
					}
					else
						mfdb_src.fd_addr = src;
				}
			}
		}
		else
			return;		
	}
	v_hide_c(vdi_handle);
	vro_cpyfm(vdi_handle, S_ONLY, pxy, &mfdb_src, &mfdb_dst);
	v_show_c(vdi_handle, 1);
	if(src != NULL)
		mem_free(src);
	if(dst != NULL)
		mem_free(dst);
}

static void gem_draw_bitmaps(struct graphics_device *dev, struct bitmap **bmps, int n, int x, int y)
{
	struct bitmap *bmp = *bmps;
	if((n < 1) || (dev == NULL) || (dev->driver_data == NULL) || !vdi_handle) 
		return;
#ifdef DEBUG_GEM
	printf("gem_draw_bitmaps n: %d, x: %d, y: %d\r\n", n, x, y);
#endif
  do
	{
		gem_draw_bitmap(dev, bmp, x, y);
		x += bmp->x;
		bmp++;
	}
  while(--n);
}

#undef APPROXIMATE_COLOR

static long gem_get_color(int search_rgb)
{
	static long count_color = 16 + 64;
	long c, save_c, rgb;
#ifdef APPROXIMATE_COLOR
	long near_c = -1, 
	int max = 255;
#endif
	short rgb_colors[3];
#ifdef DEBUG_GEM
	printf("gem_get_color rgb: 0x%06X ", search_rgb);
#endif
	if(!vdi_handle)
		return(0);
	if(bpp == 4)
	{
		for(c = 0; c < 16; c++)
		{	
			if((long)search_rgb == tab_rgb[c])
			{
#ifdef DEBUG_GEM
				printf(" => %d\r\n", c);
#endif
				return(c);
			}
		}
#ifdef DEBUG_GEM
		printf(" => not found\r\n");
#endif
		return(0);
	}
	/* search if the color exist inside the user lowest 16 colors */
	for(c = 0; c < 16; c++)
	{
		int r, g, b; 
		vq_color(vdi_handle, c, 0, rgb_colors);
		if(rgb_colors[0] == 1000)
			r = 255;
		else
			r = ((int)rgb_colors[0] << 8) / 1001;
		if(rgb_colors[1] == 1000)
			g = 255;
		else
			g = ((int)rgb_colors[1] << 8) / 1001;
		if(rgb_colors[2] == 1000)
			b = 255;
		else
			b = ((int)rgb_colors[2] << 8) / 1001;
		rgb = ((long)r << 16) + ((long)g << 8) + (long)b;
		if((long)search_rgb == rgb)
		{
#ifdef DEBUG_GEM
			printf(" => %d\r\n", c);
#endif
			return(c);
		}
#ifdef APPROXIMATE_COLOR
		if(pixelsize == 1)
		{
			int delta_r = ((search_rgb >> 16) & 255) - r;
			int delta_g = ((search_rgb >> 8) & 255) - g;
			int delta_b = (search_rgb & 255) - b;
			if(delta_r < 0)
				delta_r = -delta_r;
			if(delta_g < 0)
				delta_g = -delta_g;
			if(delta_b < 0)
				delta_b = -delta_b;
			if(delta_r > delta_g)
				delta_g = delta_r;
			if(delta_b > delta_g)
				delta_g = delta_b;
			if(delta_g < max)
			{
				max = delta_g;
				near_c = c;
			}
		}
#endif
	}
	if(pixelsize != 1)
		c += 64; /* jump png reduced palette to 64 colors in 256 colors */
	save_c = c;
	/* search if the color exist */
	for(; c < count_color; c++)
	{
		if((long)search_rgb == tab_rgb[c])
		{
#ifdef DEBUG_GEM
			printf(" => %d\r\n", c);
#endif
			return(c);
		}
#ifdef APPROXIMATE_COLOR
		if(pixelsize == 1)
		{
			int r = (tab_rgb[c] >> 16) & 255;
			int g = (tab_rgb[c] >> 8) & 255;
			int b = tab_rgb[c] & 255;
			int delta_r = ((search_rgb >> 16) & 255) - r;
			int delta_g = ((search_rgb >> 8) & 255) - g;
			int delta_b = (search_rgb & 255) - b;
			if(delta_r < 0)
				delta_r = -delta_r;
			if(delta_g < 0)
				delta_g = -delta_g;
			if(delta_b < 0)
				delta_b = -delta_b;
			if(delta_r > delta_g)
				delta_g = delta_r;
			if(delta_b > delta_g)
				delta_g = delta_b;
			if(delta_g < max)
			{
				max = delta_g;
				near_c = c;
			}
		}
#endif
	}
#ifdef APPROXIMATE_COLOR
	if((bpp == 8) && (max <= 20) && (near_c >= 0))
	{
#ifdef DEBUG_GEM
		printf(" => %d (delta: %d)\r\n", near_c, max);
#endif
		return(near_c);
	}
#endif	
	if(bpp == 8) /* 256 to 64 bmp color index */
	{
		int r = ((search_rgb >> 16) & 255) / 36;
		int g = ((search_rgb >> 8) & 255) / 36;
		int b = (search_rgb & 255) / 84;
		int index = (r << 5) + (g << 2) + b; /* color index */
		r = reduce_palette[(index >> 5) & 7] * 36;
		g = reduce_palette[(index >> 2) & 7] * 36;
		b = (index & 3) * 84;
		search_rgb = (r << 16) + (g << 8) + b;
		/* search again with a fixed search_rgb color */
		for(c = save_c; c < count_color; c++)
		{
			if((long)search_rgb == tab_rgb[c])
			{
#ifdef DEBUG_GEM
				printf(" => %d\r\n", c);
#endif
				return(c);
			}
		}
	}
	/* else define a color */
	if(count_color < 256)
	{
		tab_rgb[count_color] = (long)search_rgb;
		rgb_colors[0] = (((search_rgb >> 16) & 255) * 1001) >> 8;
		rgb_colors[1] = (((search_rgb >> 8) & 255) * 1001) >> 8;
		rgb_colors[2] = ((search_rgb & 255) * 1001) >> 8;
		vs_color(vdi_handle, count_color++, rgb_colors);
	}
	else
		c = 16 + 64;
#ifdef DEBUG_GEM
	printf(" => %d\r\n", c);
#endif
	return(c);
}

static void gem_fill_area(struct graphics_device *dev, int x1, int y1, int x2, int y2, long color)
{
	struct gem_window *view;
	short rect[4];
	if((x1 >= x2) || (y1 >= y2))
		return;
	if((dev == NULL) || (dev->driver_data == NULL) || !vdi_handle) 
		return;
	view = (struct gem_window *)dev->driver_data;
	if(view->iconified)
		return;
	if(!gem_test_win_top(dev, x1, y1, x2, y2))
		return;
#ifdef DEBUG_GEM
	printf("gem_fill_area x1: %d, y1: %d, x2: %d, y2: %d, color: %d\r\n", x1, y1, x2, y2, color);
#endif
	v_hide_c(vdi_handle);
	vsf_color(vdi_handle, color);
	vsf_interior(vdi_handle, 1);
	vsf_perimeter(vdi_handle, 0);
	vswr_mode(vdi_handle, MD_REPLACE);
	rect[0] = (short)x1 + view->x;
	rect[1] = (short)y1 + view->y;
	rect[2] = (short)x2 + view->x - 1;
	rect[3] = (short)y2 + view->y - 1;
	v_bar(vdi_handle, rect);
	v_show_c(vdi_handle, 1);
}

static void gem_draw_hline(struct graphics_device *dev, int x1, int y, int x2, long color)
{
	struct gem_window *view;
	short tab[4];
	if(x1 >= x2)
		return;
	if((dev == NULL) || (dev->driver_data == NULL) || !vdi_handle) 
		return;
	view = (struct gem_window *)dev->driver_data;
	if(view->iconified)
		return;
	if(!gem_test_win_top(dev, x1, y, x2, y))
		return;
#ifdef DEBUG_GEM
	printf("gem_draw_hline x1: %d, y: %d, x2: %d, color: %d\r\n", x1, y, x2, color);
#endif
	v_hide_c(vdi_handle);
	vsl_type(vdi_handle, 1);
	vsl_color(vdi_handle, color);
	vsl_ends(vdi_handle, 0, 0);
	vsl_width(vdi_handle, 1);
	vswr_mode(vdi_handle, MD_REPLACE);
	tab[0] = (short)x1 + view->x;
	tab[1] = tab[3] = (short)y + view->y;
	tab[2] = (short)x2 - 1 + view->x;
	v_pline(vdi_handle, 2, tab);
	v_show_c(vdi_handle, 1);
}

static void gem_draw_vline(struct graphics_device *dev, int x, int y1, int y2, long color)
{
	struct gem_window *view;
	short tab[4];
	if(y1 >= y2)
		return;
	if((dev == NULL) || (dev->driver_data == NULL)) 
		return;
	view = (struct gem_window *)dev->driver_data;
	if(view->iconified)
		return;
	if(!gem_test_win_top(dev, x, y1, x, y2))
		return;
#ifdef DEBUG_GEM
	printf("gem_draw_vline x: %d, y1: %d, y2: %d, color: %d\r\n", x, y1, y2, color);
#endif
	v_hide_c(vdi_handle);
	vsl_type(vdi_handle, 1);
	vsl_color(vdi_handle, color);
	vsl_ends(vdi_handle, 0, 0);
	vsl_width(vdi_handle, 1);
	vswr_mode(vdi_handle, MD_REPLACE);
	tab[0] = tab[2] = (short)x + view->x;
	tab[1] = (short)y1 + view->y;
	tab[3] = (short)y2 - 1 + view->y;
	v_pline(vdi_handle, 2, tab);
	v_show_c(vdi_handle, 1);
}

static int gem_hscroll(struct graphics_device *dev, struct rect_set **set, int sc)
{
	struct gem_window *view;
	MFDB mfdb_dst;
	short pxy[8];
	if((dev == NULL) || (dev->driver_data == NULL) || !vdi_handle) 
		return(0);
	view = (struct gem_window *)dev->driver_data;
	if(view->iconified)
		return(0);
  *set = NULL;
	if((dev->clip.x1 >= dev->clip.x2) || (dev->clip.y1 >= dev->clip.y2))
		return(0);
	if((sc <= dev->clip.x1 - dev->clip.x2) || (sc >= dev->clip.x2 - dev->clip.x1))
		return(1);
#ifdef DEBUG_GEM
	printf("gem_hscroll\r\n");
#endif
	v_hide_c(vdi_handle);
	pxy[0] = (short)dev->clip.x1 + view->x;
	pxy[1] = (short)dev->clip.y1 + view->y;
	pxy[2] = (short)dev->clip.x2 + view->x - 1;
	pxy[3] = (short)dev->clip.y2 + view->y - 1;
	pxy[4] = (short)dev->clip.x1 + (short)sc + view->x;
	pxy[5] = (short)dev->clip.y1 + view->y;
	pxy[6] = (short)dev->clip.x2 + (short)sc + view->x - 1;
	pxy[7] = (short)dev->clip.y2 + view->y - 1;
	memset(&mfdb_dst, 0, sizeof(MFDB));
	vro_cpyfm(vdi_handle, S_ONLY, pxy, &mfdb_dst, &mfdb_dst);
	v_show_c(vdi_handle, 1);
	return(1);
}

static int gem_vscroll(struct graphics_device *dev, struct rect_set **set, int sc)
{
	struct gem_window *view;
	MFDB mfdb_src, mfdb_dst;
	short pxy[8];
	if((dev == NULL) || (dev->driver_data == NULL) || !vdi_handle) 
		return(0);
	view = (struct gem_window *)dev->driver_data;
  *set = NULL;
	if(view->iconified)
		return(0);
	if(!sc || (dev->clip.x1 >= dev->clip.x2) || (dev->clip.y1 >= dev->clip.y2))
		return(0);
	if((sc <= dev->clip.y1 - dev->clip.y2) || (sc >= dev->clip.y2 - dev->clip.y1))
		return(1);
#ifdef DEBUG_GEM
	printf("gem_vscroll\r\n");
#endif
	v_hide_c(vdi_handle);
	pxy[0] = (short)dev->clip.x1 + view->x;
	pxy[1] = (short)dev->clip.y1 + view->y;
	pxy[2] = (short)dev->clip.x2 + view->x - 1;
	pxy[3] = (short)dev->clip.y2 + view->y - 1;
	pxy[4] = (short)dev->clip.x1 + view->x;
	pxy[5] = (short)dev->clip.y1 + (short)sc + view->y;
	pxy[6] = (short)dev->clip.x2 + view->x - 1;
	pxy[7] = (short)dev->clip.y2 + (short)sc + view->y - 1;
	memset(&mfdb_dst, 0, sizeof(MFDB));
	vro_cpyfm(vdi_handle, S_ONLY, pxy, &mfdb_dst, &mfdb_dst);
	v_show_c(vdi_handle, 1);
	return(1);
}

static void gem_set_clip_area(struct graphics_device *dev, struct rect *r)
{
	struct gem_window *view;
	if((dev == NULL) || (dev->driver_data == NULL) || !vdi_handle) 
		return;
	view = (struct gem_window *)dev->driver_data;
	if(view->iconified)
		return;
#ifdef DEBUG_GEM
	printf("gem_set_clip_area x1: %d, y1: %d, x2: %d, y2: %d\r\n", r->x1, r->y1, r->x2, r->y2);
#endif
	memcpy(&dev->clip, r, sizeof(struct rect));
	if(vdi_handle)
	{
		short rect[4];
		rect[0] = (short)r->x1 + view->x;
		rect[1] = (short)r->y1 + view->y;
		rect[2] = (short)r->x2 + view->x - 1;
		rect[3] = (short)r->y2 + view->y - 1;
		rect[2] -= (rect[0] - 1); /* x, y, w, h */
		rect[3] -= (rect[1] - 1);
		if((rect[2] <= 0) || (rect[3] <= 0)  /* bad clipping */
		 || (view->flag_redraw
		  && !rc_intersect((GRECT *)&view->redraw, (GRECT *)rect)))
		{
			view->clip.x = view->clip.y = 0;
			view->clip.w = view->clip.h = 1;
			rect[0] = 0;
			rect[1] = 0;
			rect[2] = 0;
			rect[3] = 0;
		}
		else
		{
			view->clip.x = rect[0];
			view->clip.y = rect[1];			
			view->clip.w = rect[2];
			view->clip.h = rect[3];			
			rect[2] += (rect[0] - 1);
			rect[3] += (rect[1] - 1);			
		}
		vs_clip(vdi_handle, 1, rect);
	}	
}

static void gem_set_title(struct graphics_device *dev, unsigned char *title)
{
	static unsigned char name[256];
	if(dev != NULL)
	{
		struct gem_window *view = (struct gem_window *)dev->driver_data;
		if(view != NULL)
		{
#ifdef DEBUG_GEM
			printf("gem_set_title\r\n");
#endif
			strncpy(name, title, 255);
			gem_convert_text(name);
			if(view->handle >= 0)
				wind_set(view->handle, WF_NAME, (short)(((unsigned long)name)>>16), (short)(((unsigned long)name) & 0xffff), 0, 0);
		}
	}
}

struct graphics_driver gem_driver =
{
	"gem",
	gem_init_driver,
	gem_init_device,
	gem_shutdown_device,
	gem_shutdown_driver,
	gem_get_driver_param,
	gem_get_empty_bitmap,
	/*gem_get_filled_bitmap,*/
	gem_register_bitmap,
	gem_prepare_strip,
	gem_commit_strip,
	gem_unregister_bitmap,
	gem_draw_bitmap,
	gem_draw_bitmaps,
	gem_get_color,
	gem_fill_area,
	gem_draw_hline,
	gem_draw_vline,
	gem_hscroll,
	gem_vscroll,
	gem_set_clip_area,	
	dummy_block,
	dummy_unblock,
	gem_set_title,
	NULL, /* exec */
	0,    /* depth */
	0, 0, /* size */
	0,    /* flags */
	0,    /* codepage */
	NULL, /* shell */
};

#endif /* GRDRV_GEM */
